# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from distutils.version import LooseVersion, StrictVersion
import pkg_resources

import os,re
import sys
from copy import deepcopy
import configparser

from astropy.io.fits import Header
import astropy.io.ascii as asciitable
from astropy.table import Table, Column

import math
import numpy as np
np.random.seed(seed=1234)

# LOCAL IMPORTS
from .__version__ import __version__
from .math_utils import merge_where_nan, median_clip, safe_exp

from .instruments import *
from .hyperspectral_cube import HyperspectralCube as HyperCube
from .string_stdout import StringStdOut

from .model_class import Model
from .model_sersic3d import ModelSersic
from .galaxy_parameters import GalaxyParameters, GalaxyParametersError
from .plot_utilities import Plots
from .mcmc import MCMC

#will be removed
DiskModel = ModelSersic #for backward compatibility
DefaultModel = ModelSersic

OII = {'wave': [3726.2, 3728.9], 'ratio':[0.8,1.0]}

# LOGGING CONFIGURATION
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('GalPaK')



# OPTIONAL IMPORTS
try:
    import bottleneck as bn
except ImportError:
    logger.info(" bottleneck (optional) not installed, performances will be degraded")
    import numpy as bn
try:
    import pyfftw
except ImportError:
    logger.info(" PyFFTW (optional) not installed, performances will be degraded")
try:
    import mpdaf
    logger.info("Found MPDAF version %s" % (mpdaf.__version__))
    mpdaf_there=True
except ImportError:
    mpdaf_there=False
    logger.warning(" MPDAF (optional) not installed / not required")
try:
    import emcee
    emcee_there=True
    logger.info("Found EMCEE version %s" % (emcee.__version__))
    logger.warning("EMCEE tested for version > 3.0")
except ImportError:
    emcee_there=False
    logger.warning(" EMCEE (optional) not installed / not required. So option use_emcee is disabled")
try:
    import dynesty
    dynesty_there=True
    logger.info("Found Dynesty version %s \n EXPERIMENTAL and UNSUPPORTED" % (dynesty.__version__))
except ImportError:
    dynesty_there=False
    logger.warning(" Dynesty (optional) not installed / not required. ")
try:
    import pymultinest
    multinest_there=True
    logger.info("Found PyMultinest version %s" % (pkg_resources.get_distribution("pymultinest").version))
except ImportError:
    multinest_there=False
    logger.warning(" PyMultinest (optional) not installed / not required. ")



#Python3 compatibility
try:
  basestring
except NameError:
  basestring = str

if sys.version>='2.7' and sys.version<'3.5':
    reload(sys)
    sys.setdefaultencoding('utf-8')


class GalPaK3D(Plots, MCMC):
    """
    GalPaK3D is a tool to extract Galaxy Parameters and Kinematics from
    3-Dimensional data, using reverse deconvolution with Bayesian analysis
    Markov Chain Monte Carlo. (random walk)

    cube: HyperspectralCube|string
        The actual data on which we'll work ; it should contain only one galaxy.
        Can be a HyperspectralCube object, a string filename to a FITS file, or
        even MPDAF's ``mpdaf.obj.Cube``.
    seeing: float
        Aka the Point Spread Function's Full Width Half Maximum.
        This convenience parameter, when provided, will override the FWHM value
        of the instrument's PSF.
    instrument: Instrument
        The instrument configuration to use when simulating convolution.
        The default is :class:`MUSE <galpak.MUSE>`.
    crval3: float
        A value for the cube's header's CRVAL3 when it is missing.
        You should update your cube's header.
    crpix3: float
        A value for the cube's header's CRPIX3 when it is missing.
        You should update your cube's header.
    cunit3: float
        A value for the cube's header's CUNIT3 when it is missing.
        You should update your cube's header.
    cdelt3: float
        A value for the cube's header's CDELT3 when it is missing.
        You should update your cube's header.
    cunit1: float
        A value for the cube's header's CUNIT1 (&2) when it is missing.
        You should update your cube's header.
    force_header_update: bool
        Set to True to force the update of the above header cards,
        when their values are not missing.
        Note: These will not be saved into the FITS file. (if the cube is one)

    """
    logger.info(' Running galpak ' + __version__)

    def __init__(self, cube, variance=None, model=None,
                 seeing=None, instrument=None, quiet=False,
                 crval3=None, crpix3=None, cunit3=None, cdelt3=None, ctype3=None, cunit1=None,
                 force_header_update=False):

        # DEVS : If you change the signature above,
        # remember to update the run() in api.py

       # Prepare output attributes
        self.acceptance_rate = 100.
        self.galaxy = GalaxyParameters()
        self.stdev = GalaxyParameters()
        self.chain = None
        self.sub_chain = None

        self.psf3d = None
        self.convolved_cube = None
        self.deconvolved_cube = None
        self.residuals_cube = None
        self.residuals_map = None
        self.variance_cube = None
        #true intrinsic maps
        self.true_flux_map = None
        self.true_velocity_map = None
        self.true_disp_map = None
        self.error_maps = False
        self.true_flux_map_error = None
        self.true_velocity_map_error = None
        self.true_disp_map_error = None
        #observed maps
        self.obs_flux_map = None
        self.obs_velocity_map = None
        self.obs_disp_map = None

        self.max_iterations = None
        self.method = None
        self.chain_fraction = None
        self.percentile = None
        self.initial_parameters = None
        self.min_boundaries = None
        self.max_boundaries = None
        self.known_parameters = None
        self.random_scale = None
        self.reduce_chi = None
        self.chi_stat = 'gaussian'
        self.chi_at_p = None
        self.best_chisq = None
        self.stats = None
        self.BIC = None
        self.DIC = None
        self.mcmc_method = None
        self.mcmc_sampling = None
        #self.redshift = None

        # Assign the logger to a property for convenience, and set verbosity
        self.logger = logger
        self.version = __version__
        self.config = configparser.RawConfigParser()

        self.model = None
        if quiet:
            self._set_verbose(None)
            self.verbose=None
        else:
            self._set_verbose(True)

        # Set up the input data cube
        if isinstance(cube, basestring):
            self.logger.info('Reading cube from %s' % (cube))
            cube = HyperCube.from_file(cube, verbose=not quiet)
        elif isinstance(cube, HyperCube):
            self.logger.info('Provided cube is a HyperSpectral Object')
        elif mpdaf_there:
            if isinstance(cube, mpdaf.obj.Cube):
                self.logger.info('Provided Cube is a mpdaf Cube object')
                cube = HyperCube.from_mpdaf(cube, verbose=not quiet)
            else:
                raise TypeError("Provided cube is not a HyperspectralCube "
                                    "nor mpdaf's Cube")
        else:
                raise TypeError("Provided cube is not a HyperspectralCube ")
        if cube.is_empty():
            raise ValueError("Provided cube is empty")
        self.cube = cube

        if not self.cube.has_header():
            self.logger.info("Reading hyperspectral cube without header. "
                             "Creating a minimal one.")
            self.cube.header = Header()

        # GalPaK needs a sane Cube
        #self.cube.sanitize()

        # Set up the variance
        if variance is not None:
            self.logger.info('Using user-provided variance input')
            if isinstance(variance, basestring):
                self.logger.info("Read provided variance cube   %s into HyperCube." % (variance))
                variance = HyperCube.from_file(variance)
                variance_cube = variance.data
            elif isinstance(variance, HyperCube):
                if variance.data is None:
                    self.logger.warning("Provided variance cube is empty.")
                else:
                    self.logger.info("Saving variance into varianceCube")
                    variance_cube = variance.data
            elif not (isinstance(variance,np.float) or  isinstance(variance, HyperCube)):
                raise TypeError("Provided variance is not a string nor HyperCube nor a float")
        else:
            variance_cube = self.cube.var


        # Set up the instrument's context
        if instrument is None:
            instrument = MUSE()
            self.logger.warning('Using the MUSE instrument per default. '
                                'You should specify your own instrument.')
        if isinstance(instrument, basestring):
            #read config
            instrument = self._read_instrument(instrument)
            #raise ValueError("Instrument needs to be an instance of "
            #                 "Instrument, not a string.")
        if not isinstance(instrument, Instrument):
            raise ValueError("Instrument needs to be an instance of Instrument")
        self.instrument = instrument

        ## Set default cube specs from instrument when missing from headers
        # as we don't want to rely on the input fits having properly set headers.
        # So, we aggregate in the cube our own specs (in " and µm) for our personal use :
        # - xy_step
        # - z_step
        # - z_central

        # 1. Patch up the HyperCube's missing values
        try:
            self.cube.patch(
                crval3=crval3, crpix3=crpix3, cunit3=cunit3,
                cdelt3=cdelt3, ctype3=ctype3, cunit1=cunit1, force=force_header_update
            )
        except ValueError:
            raise ValueError("The cube already has one of the header cards "
                             "you're trying to provide. "
                             "Use force_header_update=True to override.")

        self.logger.debug('Header after patch : %s' % self.cube)
        #set xy_step z_step and z_central

        # 2. Set cube metadata from the the instrument if header is incomplete
        self.cube.defaults_from_instrument(instrument=instrument)

        # 3. Initialize steps xy_steps z_steps z_cunnit and z_central
        self.cube.initialize_self_cube()

        # 4. Calibrate the instrument with the cube
        self.instrument.use_pixelsize_from_cube(self.cube)

        self.logger.debug('z central : %4.e' % (self.cube.z_central) )

        # Override the PSF FWHM (aka. seeing) if provided
        if seeing is not None and self.instrument.psf is not None:
            try:
                self.instrument.psf.fwhm = seeing
            except AttributeError:
                raise IOError("You provided a seeing but your instrument's PSF has no FWHM.")

        # Handle the variance, when provided, or generate one
        if variance_cube is not None:
            self.logger.info("Replacing 0s in the variance cube by 1e12")
            variance_cube = np.where(variance_cube == 0.0, 1e12, variance_cube)
        else:
            # Clip data, and collect standard deviation sigma
            self.logger.warning("No variance provided. Estimating Variance from edge statistics")
            clipped_data, clip_sigma, __ = median_clip(self.cube.data[:, 2:-4, 2:4], 2.5)#yband at x[2:4]
            self.logger.info("Computed stdev from the edges: sigma=%.e" % clip_sigma)
            # Adjust stdev margin if it is zero, as we'll divide with it later on
            if not np.isfinite(clip_sigma):
                clipped_data, clip_sigma, __ = median_clip(self.cube.data, 2.5)
                self.logger.info("reComputing stdev from the whole cube: sigma=%.e" % clip_sigma)

            if np.size(clip_sigma) == 1 and clip_sigma == 0:
                clip_sigma = 1e-20
            variance_cube = clip_sigma ** 2 *np.ones_like(self.cube.data)
            self.logger.info('Variance estimated is %s ' % (str(clip_sigma**2)))
        # Save the variance cube
        self.variance_cube = variance_cube   # cube of sigma^2
        self.error_cube = np.sqrt(self.variance_cube)  # cube of sigmas

        # Provide the user with some feedback
        self.logger.info("Setting up with the following setup :\n%s" % self.instrument)

        # 5. Init model
        #  Set up the model context
       #  Set up the model context
        if isinstance(model, basestring)==True:
            model = self._read_model(deepcopy(model))
        if model is not None:
            self._init_model(model)
            self.logger.info("Setting up the model : %s" % (self.model.__name__()))
            self.logger.info("Model setup :\n%s" % (self.model) )

    def _init_model(self, model):

        # Set up the simulation model
        if model is not None:
            self.model = model
            self.logger.info("Init boundaries from model '%s'" %(self.model.__name__()))
            self.model_dict = self.model.__dict__

            #important:
            self.model.pixscale = self.cube.xy_step

            # Compute a flux estimation
            # fixme: add weighted sum with variance if present
            self.flux_est = bn.nansum(self.cube.data)
            if self.flux_est < 0:
                self.logger.warning(
                    "WARNING: Initial flux (%4.2e) is <0 -- "
                    "likely wrong, will recompute it ignoring <0 values"
                    % self.flux_est
                )
                self.flux_est = np.sum(np.where(self.cube.data > 0, self.cube.data, 0))
                self.logger.warning('Initial flux is now %4.2e' % self.flux_est)
                self.logger.info('TIP: use `initial_parameters=` to set the flux')
            else:
                self.logger.info('Initial flux is %4.2e' % self.flux_est)

            # Default boundaries
            self.min_boundaries = self.model.min_boundaries(self)
            self.max_boundaries = self.model.max_boundaries(self)

            # Default initial parameters
            self.initial_parameters = (self.max_boundaries + self.min_boundaries) / 2.

            #set known parameters to ones
            self.known_parameters = self.model.Parameters()


    def run_mcmc(self, max_iterations=15000,
                 method_chain='last',
                 last_chain_fraction=60,
                 percentile=95,
                 model=None,
                 chi_stat='gaussian',
                 mcmc_method='galpak',
                 mcmc_sampling=None,
                 min_boundaries=None,
                 max_boundaries=None,
                 known_parameters=None,
                 initial_parameters=None,
                 random_scale=None,
                 min_acceptance_rate=10,
                 verbose=True,
                 emcee_nwalkers=30,
                 **kwargs):
        # DEVS : If you change the signature above,
        # remember to update the run() in api.py
        """
        Main method_chain of GalPak, computes and returns the galaxy parameters
        as a :class:`GalaxyParameters <galpak.GalaxyParameters>` object
        using reverse deconvolution with a MCMC.

        Also fills up the following attributes :
            - chain
            - psf3d
            - deconvolved_cube
            - convolved_cube
            - residuals_cube (Data-Model in units of sigma)
            - residuals_map (average of data-model in units of sigma or = 1/N_z.
                             Sum_z Residuals_cube.  sqrt(Nz) )
            - acceptance_rate
            - galaxy (same object as returned value)
                      with Vmax forced to be positive [and 180 added to PA]
            - stdev (also available as galaxy.stdev)
            - true_flux_map
            - true_velocity_map
            - true_disp_map

        Stops iteration if acceptance rate drops below ``min_acceptance_rate`` %
        or when ``max_iterations`` are reached.

        max_iterations: int
            Maximum number of useful iterations.

        method_chain: 'chi_sorted' | 'chi_min' | 'last' | 'MAP'
            Method used to determine the best parameters from the chain.
                - 'last' (default) : mean of the last_chain_fraction(%) last parameters of the chain
                - 'chi_sorted' : mean of the last_chain_fraction(%) best fit parameters of the chain
                - 'chi_min' : mean of last_chain_fraction(%) of the chain around the min chi
                - 'MAP': Parameters at Maximum At Posteriori, i.e. at chi_min

        last_chain_fraction: int
            Last Chain fraction (in %) used to compute the best parameters.
            Defaults to 60.

        model = DefaultModel()
            see class DiskModel or ModelSersic

        chi_stat: 'gaussian' [default] | 'Mighell' | 'Neyman' | 'Cstat' | 'Pearson'
            The chi2 statitics
            https://heasarc.gsfc.nasa.gov/xanadu/xspec/manual/XSappendixStatistics.html
                - 'gaussian' (default): Sum (D - M)^2 / e
                - 'Neyman'  Sum (D - M )^2 / max(D,1)
                - 'Mighell' Sum (D + min(D,1) - M)^2 / (D+1)  Mighell http://adsabs.harvard.edu/abs/1999ApJ...518..380M
                - 'Cstat' Sum ( M - D + D * log(D/M) ) Cash statistique Humphrey 2009, http://adsabs.harvard.edu/abs/2009ApJ...693..822H
                - 'Pearson'  Sum ( M - D )^2 / M  Pearson statistic  Humphrey 2009, http://adsabs.harvard.edu/abs/2009ApJ...693..822H


        mcmc_method: 'galpak' [default] | 'emcee_walkers'| 'emcee_MH' | 'dynesty' | 'multinest'
            The MCMC method.
                - galpak:   for the original MCMC algorithm using Cauchy proposal distribution
                - emcee_MH: emcee Metropolis Hasting
                - emcee_walkers: emcee multi-Walkers algorithms with Moves if version>=3.0
                - dynesty: unsupported
                - multinest: using Importance Nested Sampling w/ pyMultinest
                - pymc3: to be implemented

        mcmc_sampling: None [default] | 'Cauchy'  | 'Gaussian' | 'DE' | 'walkers'
            The sampling proposal distribution
            - 'Cauchy' default when mcmc_method = 'galpak' or 'emcee_MH'
            - 'walkers' (=StretchMove) default when mcmc_method ='emcee_walkers'

        min_boundaries: ndarray|GalaxyParameters
            The galaxy parameters will never be less than these values.
            Will override the default minimum boundaries for the parameters.
            If any of these values are NaN, they will be replaced by the default ones.

        max_boundaries: ndarray|GalaxyParameters
            The galaxy parameters will never be more than these values.
            Will override the default minimum boundaries for the parameters.
            If any of these values are NaN, they will be replaced by the default ones.

        known_parameters: ndarray|GalaxyParameters
            All set parameters in this array will be skipped in the MCMC,
            the algorithm will not try to guess them.

        initial_parameters: ndarray|GalaxyParameters
            The initial galaxy parameters of the MCMC chain.
            If None, will use the inital parameters provided by the model.
            The galaxy parameters not initialized by the model or by this
            parameter will be set to the mean of the boundaries.


        random_scale: float
            Scale the amplitude of the MCMC sampling by these values.
            This is an important parameter to adjust for reasonable acceptance rate.
            The acceptance rate should be around 30-50%.
            If the acceptance rate is <20-30% (too low), decrease random_scale
            IF the acceptance rate is >50-60% (too high), increase random_scale

        verbose: boolean
            Set to True to output a detailed log of the process.
            The run is faster when this is left to False.

        """
        # DEVS : If you change the signature above,
        # remember to update the run() in api.py

        #re initialize psf
        self.instrument.psf3d_fft = None

        # Save the parameters (the animation uses them)
        self.max_iterations = max_iterations
        self.method = method_chain
        self.chain_fraction = last_chain_fraction
        self.chi_stat = chi_stat
        self.percentile = percentile
        self.verbose = verbose

        #save intermediate model maps
        self.error_maps = None #will be set later
        self.chain_flux_map = []
        self.chain_velocity_map = []
        self.chain_dispersion_map = []

        # Set up the simulation model
        if model is None:
            if self.model is not None:
                self.logger.warning("Model already specified: '%s'" % (self.model.__name__()))
            else:
                self.model = DefaultModel()
                self.logger.warning("Will use default model '%s'" %(self.model.__name__()))
                self.logger.info("Model setup :\n%s" % (self.model) )

        elif model is not None:
            if isinstance(model, basestring):
                #read from config file
                self.model = self._read_model(model)
                self.logger.info("Model set to %s from file: %s" % (self.model.__name__(), self.model) )
            elif isinstance(model, Model):
                self.logger.warning("Model was already set")
                self.model = model
                self.logger.info("Model set to %s : %s" % (self.model.__name__(), self.model))
            else:
                raise ValueError
        self._init_model(self.model)

        #For backwards compatibility:
        ## will be removed in the future
        #if self.model.line is None:
        #    self.model.line = self.line
        #else:
        #    self.line = self.model.line
        #save attribute
        #For backwards compatibility:
        ## will be removed in the future
        ## if not None, computes Mdyn(Re)
        #if self.model.redshift is not None:
        #    self.redshift = self.model.redshift


        # Sanitize data and set arbitrary big stdev where NaNs are
        #cube_data = self.cube.data
        #cube_data = np.nan_to_num(cube_data)

        # Set verbosity
        self._set_verbose(verbose)


        dim_p = np.size(GalaxyParameters())

        # In fraction of boundary space, an arbitrarily
        # small value for closeness to boundaries
        self.eps = 0.003


        # Merge provided boundaries (if any) with default boundaries
        if isinstance(min_boundaries, basestring):
            min_boundaries = self._read_params(deepcopy(min_boundaries),'MIN')
        if isinstance(max_boundaries, basestring):
            max_boundaries = self._read_params(deepcopy(max_boundaries),'MAX')

        if min_boundaries is not None:
            min_boundaries = deepcopy(min_boundaries)
            merge_where_nan(min_boundaries, self.min_boundaries) #this returns a ndarray
            self.min_boundaries = GalaxyParameters().from_ndarray(min_boundaries)
        if max_boundaries is not None:
            max_boundaries = deepcopy(max_boundaries)
            merge_where_nan(max_boundaries, self.max_boundaries) #this returns a ndarray
            self.max_boundaries = GalaxyParameters().from_ndarray(max_boundaries)


        bug_boundaries = self.min_boundaries > self.max_boundaries
        if bug_boundaries.any():
            self.logger.debug("Min Boundaries : %s", self.min_boundaries)
            self.logger.debug("Max Boundaries : %s", self.max_boundaries)
            raise ValueError("Boundaries are WRONG, because min > max")

        # Default initial parameters
        self.initial_parameters = self.model.initial_parameters(self)

        # Create initial galaxy parameters using mean and provided values
        mean_parameters = (self.max_boundaries + self.min_boundaries) / 2.
        #complete default values with mean parameters
        merge_where_nan(self.initial_parameters, mean_parameters)
        self.logger.info("Default Param_init : %s", self.initial_parameters)

        # read initial parameter from config file
        if isinstance(initial_parameters, basestring):
            initial_parameters = self._read_params(deepcopy(initial_parameters), 'INIT')

        # Merge provided initial parameters (if any) with the defaults
        if initial_parameters is not None:
            #complete input with default values
            template = self.model.Parameters()
            merge_where_nan(template, initial_parameters)
            merge_where_nan(template, self.initial_parameters)
            self.initial_parameters = template

        self.logger.info("Initial parameters : %s", self.initial_parameters)

        # By default, try to guess all parameters
        should_guess_flags = np.ones(dim_p)  # 0: we know it / 1: try to guess

        #if using input image
        # @fixme; this is unused
        if self.model_dict['flux_profile'] == 'user' and known_parameters is None:
                raise self.logger.error(
              "With an input image it is advised to freeze "
              "the `inclination`, using `known_parameters=`.")

        # Flag parameters that we manually specified and don't need to guess
        if isinstance(known_parameters, basestring):
            known_parameters = self._read_params(deepcopy(known_parameters), 'KNOWN')

        if known_parameters is not None:
            if len(known_parameters) != dim_p:
                raise ValueError("The `known_parameters=` must be an array of "
                                 "length %d, or even better an instance of `%s`"
                                 % (dim_p, self.model.parameters_class()))
            else:
                merge_where_nan(self.known_parameters, known_parameters)

        self.logger.info("Using known parameters: %s", self.known_parameters)

        # Freeze the known parameters by flagging them as not-to-guess
        for idx in range(dim_p):
            parameter = self.known_parameters[idx]
            if not math.isnan(parameter):
                should_guess_flags[idx] = 0
                sign = math.copysign(1., parameter)
                self.min_boundaries[idx] = parameter * (1 - self.eps * sign) - self.eps
                self.max_boundaries[idx] = parameter * (1 + self.eps * sign) + self.eps
                self.initial_parameters[idx] = parameter

        # The setup is finished, let's dump some information
        self.logger.info("Min Boundary: %s", self.min_boundaries)
        self.logger.info("Max Boundary: %s", self.max_boundaries)

        # The setup is done, we can now start the MCMC loop.
        if isinstance(random_scale, basestring):
            random_scale = self._read_params(deepcopy(random_scale), 'RSCALE')

        self.random_scale = random_scale
        random_amplitude = self._init_sampling_scale(random_scale, should_guess_flags)

        # Zero random amplitude where parameters are known
        random_amplitude = random_amplitude * should_guess_flags
        self.random_amplitude = random_amplitude


        self.logger.info("Starting with χ² = %f", self.compute_chi(self.initial_parameters) / self.Ndegree)

        ## The actual MCMC #####################################################
        #self.error_maps = save_error_maps #save intermediate model maps

        #set walkers
        if mcmc_method == 'galpak' and mcmc_sampling is None:
            mcmc_sampling = 'Cauchy'
        #if mcmc_method == 'emcee_MH' and mcmc_sampling is None:
        #    mcmc_sampling = 'Cauchy'
        if mcmc_method == 'emcee_walkers' and mcmc_sampling is None:
            mcmc_sampling = 'walkers'

        self.mcmc_method = mcmc_method
        self.mcmc_sampling = mcmc_sampling


        #burnin = np.int(0.15*self.max_iterations)
        if mcmc_method == 'galpak':
            chain = self.myMCMC(max_iterations, random_amplitude, sampling_method=mcmc_sampling, min_acceptance_rate=min_acceptance_rate)
        elif mcmc_method == 'emcee_MH' and emcee_there:
            #DEFAULT EMCEE parameters for EnsembleSampler(EMCEE)
            pass

        elif mcmc_method == 'emcee_walkers' and emcee_there:
            #DEFAULT EMCEE parameters for EnsembleSampler(EMCEE)

            #kwargs_sampler = {
            #        'pool':None,  \
            #        'backend':None,\
            #        'vectorize':False,\
            #        'blobs_dtype':None,\
            #        'postargs':None,\
            #        'threads': 1,
            #        }
            emcee_threads = 4
            kwargs_emcee ={
                     'store': True, \
                     'tune': True, \
                    'thin': 30
                    }
            #update default parameters
            kwargs_emcee.update(kwargs)

            #@fixme: need to accept parallelize
            pos0 = np.array([self.initial_parameters * (1+1e-3*np.random.randn(dim_p)) for i in range(emcee_nwalkers) ])
            self.logger.critical("Running EMCEE with %d walkers on %d iterations" % (emcee_nwalkers, self.max_iterations))

            if LooseVersion(emcee.__version__)<LooseVersion('3.0'):
                raise Exception("EMCEE version not supported ", emcee.__version__)

            #EMCE version3
            if mcmc_sampling == 'Cauchy':
                from .mcmc import CauchyMove
                myMove=CauchyMove(self.random_amplitude.as_vector()**2)
                self.logger.critical("Running EMCEE Walkers with Cauchy Sampling")
            elif mcmc_sampling == 'Normal':
                from emcee.moves import GaussianMove
                myMove=GaussianMove(self.random_amplitude.as_vector()**2)
                self.logger.info("Random Ampl : %s", self.random_amplitude.as_vector())
                self.logger.critical("Running EMCEE Walkers with Gaussian Sampling")
            elif mcmc_sampling == 'DE':
                from emcee.moves import DEMove
                myMove=DEMove()
                self.logger.critical("Running EMCEE Walkers with DE Sampling")
            elif mcmc_sampling == 'Snooker':
                from emcee.moves import DESnookerMove
                myMove=DESnookerMove()
                self.logger.critical("Running EMCEE Walkers with Snooker Sampling")
            elif mcmc_sampling == 'walkers':
                myMove = None #default StretchMove from EMCEE
                self.logger.critical("Running EMCEE Walkers with default Stretch Sampling")
            elif mcmc_sampling == 'walkersCauchy':
                from .mcmc import CauchyMove
                #50/50 StretchMove and CauchyMove
                myMove = [  (emcee.moves.StretchMove(),0.6), (CauchyMove(self.random_amplitude.as_vector()**2),0.4)]
                self.logger.critical("Running EMCEE Walkers with 60/40 StretchMove() & Cauchy Sampling")
            else:
                raise Exception("mcmc_sampling not valid. Options are  ", self.SAMPLING_VALID)
            kwargs_emcee.update(kwargs)
            #Multiprocessing
            #try:
            #    from multiprocessing import Pool
            #    self.logger.info("Running EMCEE with multiprocessing")
            #    with Pool() as pool:
            #        self.sampler = emcee.EnsembleSampler(emcee_nwalkers, dim_p, self, moves=myMove, pool=pool)
            #        if self.verbose is not True:
            #            self.sampler.run_mcmc(pos0, self.max_iterations, progress=True, **kwargs_emcee)
            #        else:
            #            for state in self.sampler.sample(pos0, iterations=self.max_iterations, **kwargs_emcee):
            #                for k, r in enumerate(state.coords):
            #                    print("%d %s log L=%f" % (self.sampler.iteration, self.model.Parameters().from_ndarray(r), \
            #                                              state.log_prob[k]))


            #except ImportError:
            self.logger.info(" Running EMCEE with 4 threads")
            self.sampler = emcee.EnsembleSampler(emcee_nwalkers, dim_p, self, moves=myMove, threads=4)


            if self.verbose is not True:
                self.sampler.run_mcmc(pos0, self.max_iterations, progress=True, **kwargs_emcee)
            else:
                for state in self.sampler.sample(pos0, iterations=self.max_iterations, **kwargs_emcee):
                    for k,r in enumerate(state.coords):
                        print("%d %s log L=%f" %(self.sampler.iteration, self.model.Parameters().from_ndarray(r),\
                                             state.log_prob[k]))

            self.sampler.__dict__['kwargs'] = kwargs_emcee

            self.acceptance_rate = self.sampler.acceptance_fraction
            self.logger.info("EMCEE MH: Acceptance: %s " % (self.sampler.acceptance_fraction))
            #self.logger.info("EMCEE MH: Naccepted states ",  (self.sampler.naccepted))
            #flat_chain = self.sampler.get_chain(discard=burnin, flat=True)
            chain_data = self.sampler.flatchain
            chain = Table(chain_data, names=self.model.Parameters().names)
            lnprob = self.sampler.flatlnprobability
            chain.add_column(Column(-2*lnprob / self.Ndegree), name='reduced_chi')
        elif mcmc_method == 'dynesty' and dynesty_there:
            # "Dynamic" nested sampling.
            #
            nlive = 500
            self.sampler = dynesty.DynamicNestedSampler(self.loglike, self.ptform, dim_p \
                                                        , bound='single' #to force posterior weights
                                                        , sample='unif' #unif/hscale
                                                        )
            kwargs_dynesty = {'nlive_init': 30 , 'nlive_batch': 200 }
            kwargs_dynesty.update(kwargs)

            self.logger.critical('EXPERIMENTAL Running Dynesty with ', kwargs_dynesty)
            self.sampler.run_nested(wt_kwargs={'pfrac': 0.9} #posterior based
                                    , maxiter = self.max_iterations
                                    , **kwargs_dynesty
                                    )

            self.dresults = self.sampler.results

            chain_data = self.dresults.samples
            chain = Table(chain_data, names=self.galaxy.names)
            lnprob = self.dresults.logz
            chain.add_column(Column(-2*lnprob / self.Ndegree), name='reduced_chi')

        elif mcmc_method == 'multinest':
            """
                run(LogLikelihood, Prior, n_dims, n_params=None, n_clustering_params=None,
                wrapped_params=None, importance_nested_sampling=True, multimodal=True,
                const_efficiency_mode=False, n_live_points=400, evidence_tolerance=0.5,
                sampling_efficiency=0.8, n_iter_before_update=100, null_log_evidence=-1e+90,
                max_modes=100, mode_tolerance=-1e+90, outputfiles_basename=u'chains/1-',
                seed=-1, verbose=False, resume=True, context=0, write_output=True,
                log_zero=-1e+100, max_iter=0, init_MPI=True, dump_callback=None)
            """

            outpath = './pymulti'
            if os.path.isdir(outpath) is False:
                os.mkdir(outpath)
            output = outpath + '/out'

            #default parameters
            kwargs_multi={'n_live_points': 200, \
                        'evidence_tolerance':0.5, \
                        'n_iter_before_update' : 200, \
                        'const_efficiency_mode' : False, \
                        'sampling_efficiency':0.8, \
                        'resume' : False}
            kwargs_multi.update(kwargs)

            self.logger.critical("Running MultiNest with ", kwargs_multi)
            self.logger.info(" Multinest, ignoring max_iteration")

            pymultinest.run(self.pyloglike, self.pycube, n_dims= dim_p, \
                max_iter=0, verbose=self.verbose, \
                outputfiles_basename=output, **kwargs_multi)

            # create analyzer object
            #embedded in solve
            analyzer = pymultinest.Analyzer(dim_p, outputfiles_basename = output)


            # get a dictionary containing information about
            #   the logZ and its errors
            #   the individual modes and their parameters
            #   quantiles of the parameter posteriors
            data = analyzer.get_data()[:,:-1]
            #print(data.shape)

            stats = analyzer.get_mode_stats()

            #lnZ = stats['evidence']

            #  iterate through the "posterior chain"
            #for params in a.get_equal_weighted_posterior():
            #        print(params)

            samples = analyzer.get_equal_weighted_posterior()
            #print(chain_data.shape)

            chain = Table(samples[:,:-1], names=self.galaxy.names)

            lnprob = samples[:,-1]

            chain.add_column(Column(-2*lnprob / self.Ndegree), name='reduced_chi')

            # get the best fit (highest likelihood) point
            #bestfit_params = stats['modes'][0]['mean']
            #bestfit_params = stats['modes'][0]['maximum']
            #bestfit_params = stats['modes'][0]['maximum a posterior']
            #OR
            #bestfit_params = stats.get_best_fit()['parameters']

            self.sampler = dict(samples = samples, stats=stats, \
                kwargs=kwargs_multi
                )



        elif mcmc_method == 'pymc3':
            raise NotImplementedError

        elif mcmc_method == 'pynuts':
            raise NotImplementedError
        else:
            raise Exception("method_mcmc %s not valid. Used of of %s" % (mcmc_method, self.MCMC_VALID))

        # Store chain
        self.logger.info("self.chain : full Markov chain")
        #good_idx = np.where(chain['reduced_chi']!=0)
        #self.chain = Table(chain[good_idx])
        # Sanitize the chain if vel <0
        self.model.sanitize_chain(chain)
        self.chain = chain

        # Store PSF 3D, which may not be defined
        try:
            self.psf3d = HyperCube(self.instrument.psf3d)
        except AttributeError:
            pass

        # Extract Galaxy Parameters from chain, and store them
        self.logger.info("Extracting best parameters (medians) from chain")
        self.best_parameters_from_chain(method_chain, last_fraction=last_chain_fraction, percentile=percentile)

        # Create output cubes
        self.logger.info("self.convolved_cube : simulated convolved cube from found galaxy parameters")
        self.convolved_cube = self.create_convolved_cube(self.galaxy, self.cube.shape)
        self.convolved_cube.header = self.cube.header
        self.logger.info("self.deconvolved_cube : deconvolved cube from found galaxy parameters")
        self.deconvolved_cube = self.create_clean_cube(self.galaxy, self.cube.shape, final=True)
        self.deconvolved_cube.header = self.cube.header
        self.logger.info("self.residuals_cube : diff between actual data and convolved cube, scaled by stdev margin")
        self.residuals_cube = (self.cube - self.convolved_cube) / self.error_cube  # * np.mean(variance_cube)
        self.residuals_cube.header = self.cube.header
        #compute observed maps
        #_ = self._make_moment_maps(self.convolved_cube, mask=True) # make moment maps with cube convolved with 3DPSF
        _ = self._make_maps_Epinat(self.convolved_cube, mask=True)
        # Average of residuals, normalized to sigma_mu
        nz = self.residuals_cube.shape[0]
        self.residuals_map = (self.residuals_cube.data.sum(0) / nz) * np.sqrt(nz)

        # Compute the χ²
        self.compute_stats()
        self.logger.info("χ² at best param: %f", self.chi_at_p)
        self.logger.info("Best min χ², %f ", self.best_chisq)

        self.logger.info("BIC (full) : %f ", self.BIC)
        self.logger.info("DIC : %f ", self.DIC)


        # Show a plot of the chain if verbose, to draw attention to the chain
        if verbose:
            # Sometimes, when the number of iterations is low, plotting fails.
            # It is a complex issue with matplotlib, so we're 'try'-wrapping it.
            try:
                self.plot_mcmc(adapt_range='5stdev')
            except:
                pass

            try:
                self.plot_geweke()
            except:
                pass

        return self.galaxy


    def best_parameters_from_chain(self, method_chain='last', last_fraction=60, percentile=95):
        """
        Computes best fit galaxy parameters from chain, using medians from a specified method_chain.

        method_chain: string 'last' | 'chi_sorted' | 'chi_min' | 'MAP'
            The method to use to extract the fittest parameters from the chain.
            'last' (default) : mean of the last_fraction(%) last parameters of the chain
            'chi_sorted' : mean of the last_fraction(%) best fit parameters of the chain
            'chi_min' : mean of last_fraction(%) of the chain around the min chi
            'MAP': Parameters at Maximum At Posteriori, i.e. at chi_min
        last_fraction: float % (60 as default)
            Fraction of the end of the chain used in determining the parameters.
        percentile: float % (95 as default)
            None: the method to use to compute the errors on the parameter is the standard deviation of the median
            float: the percentile (the 68th, or 95th percentile) to be used for the errors on the parameters
            #fixme: in which case returns the lower and upper values.

        Returns the galaxy parameters and the stdev
        """

        self.method = method_chain
        self.chain_fraction = last_fraction
        self.percentile = percentile

        if self.chain is None:
            raise RuntimeError("No chain! Run .run_mcmc() first.")

        # Data correction for Vmax
        #vmax_sign = (self.chain['maximum_velocity'] < 0)
        #pa_correction = np.where(vmax_sign, self.chain['pa'] + 180., self.chain['pa'])
        #pa_correction = np.where(pa_correction > 180, pa_correction - 360, pa_correction)
        #self.chain['pa'] = pa_correction
        #self.chain['maximum_velocity'] = np.abs(self.chain['maximum_velocity'])

        cols = [ Column(data=np.cos(np.radians(self.chain['pa'])),  name='cospa'),
                 Column(data=np.cos(np.radians(self.chain['pa']*2)),name='cos2pa'),
                 Column(data=np.sin(np.radians(self.chain['pa'])),  name='sinpa'),
                 Column(data=np.sin(np.radians(self.chain['pa']*2)),name='sin2pa')
                 ]
        chain_full = self.chain.copy()
        chain_full.add_columns(list(cols))


        #extract subchain
        chain_size = np.size(chain_full)
        n = np.int(chain_size * last_fraction / 100.)  # number of samples (last_fraction(%) of total)

        idx = chain_full.argsort('reduced_chi')
        self.chain.idxsorted = idx

        if method_chain == 'chi_min' or method_chain == 'MAP':
            min_chi_index = self._get_min_chi_index()
            xmin = np.max([0, min_chi_index - n // 2])
            xmax = np.min([chain_size, min_chi_index + n // 2])
            sub_chain = chain_full[xmin: xmax]
        elif method_chain == 'last':
            sub_chain = chain_full[-n:]
            xmin = chain_size - n
            xmax = chain_size
        elif method_chain == 'chi_sorted':
            sub_chain = chain_full[idx][:n]
            sub_idx = idx[:n]
            xmin = 0
            xmax = n
        else:
            raise ValueError("Unsupported  `method_chain` '%s'"
                             % method_chain)
        #compute error model maps
        #@fixme
        # if self.error_maps:
        #    if method_chain == 'chi_min' or method_chain =='MAP':
        #        tmp_f = np.array(self.chain_flux_map)[min_chi_index - n // 2: min_chi_index + n // 2]
        #        tmp_v = np.array(self.chain_velocity_map)[min_chi_index - n // 2: min_chi_index + n // 2]
        #        tmp_s = np.array(self.chain_dispersion_map)[min_chi_index - n // 2: min_chi_index + n // 2]
        #    elif method_chain == 'last':
        #        tmp_f = np.array(self.chain_flux_map)[-n:]
        #        tmp_v = np.array(self.chain_velocity_map)[-n:]
        #        tmp_s = np.array(self.chain_dispersion_map)[-n:]
        #    elif method_chain == 'chi_sorted':
        #        tmp_f = np.array(self.chain_flux_map)[sub_idx]
        #        tmp_v = np.array(self.chain_velocity_map)[sub_idx]
        #        tmp_s = np.array(self.chain_dispersion_map)[sub_idx]
        #    self.true_flux_map_error = HyperCube(
        #        np.percentile(tmp_f, 50 + percentile/2., axis=0) -
        #        np.percentile(tmp_f, 50 - percentile/2., axis=0)
        #    )
        #    self.true_velocity_map_error = HyperCube(
        #        np.percentile(tmp_v, 50 + percentile/2., axis=0) -
        #        np.percentile(tmp_v, 50 - percentile/2., axis=0)
        #    )
        #    self.true_disp_map_error = HyperCube(
        #        np.percentile(tmp_s, 50 + percentile/2., axis=0) -
        #        np.percentile(tmp_s, 50 - percentile/2., axis=0)
        #    )

        # Compute best_parameters
        parameter_names = list(self.model.Parameters().names)

        if self.method == 'MAP':
            tmp_chain = np.array(chain_full[parameter_names].as_array().tolist())
            best_parameter = tmp_chain[idx[0]]
        else:
            tmp_chain = np.array(sub_chain[parameter_names].as_array().tolist()) #convert astropy Table into array
            best_parameter = np.median(tmp_chain, axis=0)
        # Compute errors to parameters
        sigma_parameter = np.std(tmp_chain, axis=0)

        #handle PA edges at +/-180
        #following https://ncss-wpengine.netdna-ssl.com/wp-content/themes/ncss/pdf/Procedures/NCSS/Circular_Data_Analysis.pdf
        cos1=np.median(sub_chain['cospa'])
        sin1=np.median(sub_chain['sinpa'])
        invtan = np.arctan2(sin1,cos1)
        pa_circular_best = np.degrees(invtan) # over -180;180
        #save
        pa_idx = sub_chain[parameter_names].index_column('pa')
        best_parameter[pa_idx] = np.where(pa_circular_best>0, pa_circular_best, pa_circular_best+360)

        r1 = np.sqrt(sub_chain['cospa'].sum()**2+sub_chain['sinpa'].sum()**2)
        r1 = r1/np.size(sub_chain)
        pa_circular_std = np.sqrt(-2.*np.log(r1))

        cos2=np.median(sub_chain['cos2pa'])
        sin2=np.median(sub_chain['sin2pa'])
        #r2 = np.sqrt(sub_chain['cos2pa']**2+sub_chain['sin2pa']**2)
        #r2 = np.median(r2)
        invtan2= np.arctan2(sin2,cos2)
        angle2= np.degrees(invtan2)

        ## this is??
        #pa_circular_dispersion = (1-angle2)/(2*r1**2)
        #print "pa disp  %.3f " % (pa_circular_dispersion)

        #center chain['pa']
        sub_chain_pa = sub_chain['pa'] - best_parameter[pa_idx]
        #sub_chain_pa = np.where(sub_chain_pa > 180, sub_chain_pa - 360, sub_chain_pa)
        #sub_chain_pa = np.where(sub_chain_pa < -180, sub_chain_pa + 360, sub_chain_pa)

        std_pa = np.std(sub_chain_pa)
        #This is identical to pa_circular_std
        #print "pa sigma %.3f" % (std_pa)
        #print "pa stdev %.3f " % (np.degrees(pa_circular_std))
        sigma_parameter[pa_idx] = std_pa

        #add back centering offset
        sub_chain['pa'] = sub_chain_pa + best_parameter[pa_idx]
        #keep full set
        chain_names = deepcopy(parameter_names)
        chain_names.append('reduced_chi')
        self.sub_chain = sub_chain[chain_names]
        self.chain = chain_full[chain_names]
        self.chain.xmin = xmin
        self.chain.xmax = xmax
        # self.sub_chain = sub_chain[parameter_names] #will be used for correlation plots

        #min in sub_chain:
        #self.best_chisq = np.min(self.sub_chain['reduced_chi'])
        #min absolute
        self.best_chisq = self.chain[idx[0]]['reduced_chi']

        #Save
        self.galaxy = GalaxyParameters.from_ndarray(best_parameter)
        self.galaxy.stdev = GalaxyParametersError.from_ndarray(sigma_parameter)
        self.stdev = self.galaxy.stdev

        # Compute percentiles
        if percentile is not None:
            self.logger.info('Setting %2d percentiles ' % (percentile))

            error_parameter_upper = np.percentile(self.sub_chain[parameter_names].as_array().tolist(), 50. + percentile/2., axis=0)
            error_parameter_lower = np.percentile(self.sub_chain[parameter_names].as_array().tolist(), 50. - percentile/2., axis=0)

            self.galaxy.upper = GalaxyParametersError.from_ndarray(error_parameter_upper)
            self.galaxy.lower = GalaxyParametersError.from_ndarray(error_parameter_lower)
            self.galaxy.ICpercentile = percentile



        # Store galaxy parameters and stdev
        self.logger.info("self.galaxy : fittest parameters : %s", repr(self.galaxy))
        self.logger.info("self.stdev : parameters stdev : %s", str(self.stdev))

        return None

    def compute_stats(self, snr_min=0.02):
        """
        snr_min : float [default = 0.02]
            minimum snr to compute BIC restricted over pixels with snr > snr_min
        compute stats (BIC, DIC, AIC)
        """

        dim_data = np.size(self.cube.data)

        self.chi_at_p = self.compute_chi(self.galaxy) / self.Ndegree

        self.convolved_cube = self.create_convolved_cube(self.galaxy, self.cube.shape)
        snr = (self.convolved_cube.data) / np.median(self.variance_cube)**0.5
        good = np.ones_like(snr)
        good[snr<snr_min] = 0
        dim_good = good.sum() #count only pixels with snr >snr_min and above

        Nd_good = (dim_good - self.dim_p_free - 1) # degree of freedom


        #chi2 computed over smaller volume
        #chi_at_p_r = compute_chi(self.galaxy, good_v)  / Nd_good
        variance_cube_save = deepcopy(self.variance_cube)
        self.variance_cube = self.variance_cube / good
        self.variance_cube = variance_cube_save

        # like = -0.5*np.nansum(self.variance_chi)-0.5*self.compute_chi(params)
        #BIC = -2 * like(theta)
        self.BIC = self.chi_at_p * self.Ndegree + self.dim_p_free * np.log(dim_data)


        #AIC = -2 * like + 2 dim_p
        self.AIC = self.chi_at_p * self.Ndegree + 2* self.dim_p_free

        #DIC
        if self.method != 'chi_sorted':
            log_Lp = -0.5 * self.sub_chain['reduced_chi'] * self.Ndegree
            if np.isfinite(log_Lp).all():
                pD = 2 * np.var(log_Lp)
            else:
                pD = 2 * np.nanvar(log_Lp[np.isfinite(log_Lp)==True])
            # P = 2 * (logp_max-np.mean(self.lnp))
            #pD = 2 * -0.5 * (self.chi_at_p - np.mean(self.sub_chain['reduced_chi'])) * self.Ndegree

            #DIC = -2 * like + 2 pd
            self.DIC = self.chi_at_p * self.Ndegree + 2 * pD
        else:
            self.DIC = 0
            pD=0

        self.stats = Table(np.array(['%.8f' % (self.best_chisq), '%.8f' % (self.chi_at_p),
                                     int(self.BIC), self.Ndegree,
                                     int(self.AIC), self.dim_p_free,
                                     '%.2f' % (pD), int(self.DIC),
                                     np.max(snr)]
                                    ), \
                           names=['best_chi2', 'chi2_at_p', 'BIC', 'Ndegree', \
                                  'AIC', 'k', \
                                  'pD', 'DIC', 'SNRmax'])

        if self.mcmc_method is 'multinest':
            evidence = self.sampler['stats']['global evidence']
            self.stats.add_column(evidence*-2,name='log Z')

        return self.stats

    def create_clean_cube(self, galaxy, shape, final=False):
        """
        Creates a cube containing a clean simulation of a galaxy according to
        the provided model.

        galaxy: GalaxyParameters
            The parameters upon which the simulated galaxy will be built.
        shape: Tuple of 3
            The 3D (z, y, x) shape of the resulting cube.
            Eg: (21, 21, 21)

        Returns a HyperspectralCube
        """
        # Create radial velocities and radial dispersions
        #flux_cube, vz, vz_map, s_map, sig_map, sigz_disk_map, sig_intr = \
        #    self.model._compute_galaxy_model(galaxy, shape)

        ## normalize to real flux
        #flux_map = flux_cube.sum(0)
        #flux_map = galaxy.flux * flux_map / flux_map.sum()

        if self.model is None:
                raise AssertionError(" Model is undefined. Please define model first")

        modelcube, flux_map, vz_map, s_map = self.model._create_cube(galaxy, shape,\
                                self.instrument.z_step_kms, zo=galaxy['z'])

        #@fixme: currently records all calculations..
        # if self.error_maps:
        #    self.chain_flux_map.append(flux_map)
        #    self.chain_velocity_map.append(vz_map)
        #    self.chain_dispersion_map.append(s_map)

        # This is too expensive ! We create Cubes on each iteration...
        # There are ways to optimize this, as only the last one is used.
        if final is True:
            self.true_flux_map = HyperCube(flux_map)
            self.true_velocity_map = HyperCube(vz_map)
            self.true_disp_map = HyperCube(s_map)

        #modelcube = self.model._create_cube(shape, flux_map, vz_map, s_map,
        #                         self.instrument.z_step_kms, zo=galaxy.z)
        #is this used?
        modelcube.xy_step = self.cube.xy_step
        modelcube.z_step = self.cube.z_step
        modelcube.z_central = self.cube.z_central

        return modelcube

    def create_convolved_cube(self, galaxy, shape):
        """
        Creates a cube containing a convolved simulation of a galaxy according
        to the provided model.
        The convolution is done by the instrument you provided upon
        instantiation of this class.

        galaxy: GalaxyParameters
            The parameters upon which the simulated galaxy will be built.
        shape: Tuple of 3
            The 3D (Z, Y, X) shape of the resulting cube.
            Eg: (21, 21, 21)

        Returns a HyperspectralCube
        """

        clean_cube = self.create_clean_cube(galaxy, shape)
        return self.instrument.convolve(clean_cube)

    def import_chain(self, filepath, compute_best_params=False, method_chain='last'):
        """
        Imports the chain stored in a .dat file so that you may plot.
        compute_best_parameters False[default] | True
            if True will use 'last' method and 60%
            use best_parameters_from_chain method to customize
        """

        with open(filepath, 'r') as chain_data:
            self.chain = asciitable.read(chain_data.read(), Reader=asciitable.FixedWidth)

        self.model.sanitize_chain(self.chain)

        if compute_best_params is True:
            self.best_parameters_from_chain(method_chain=method_chain, last_fraction=60, percentile=95)

    NO_CHAIN_ERROR = "No chain to plot! Run .run_mcmc() or .import_chain() " \
                     "first."


    def save(self, name, overwrite=False):
        """
        Saves the results of the MCMC to files :

        - <name>_galaxy_parameters.txt
            A plain text representation of the parameters of the galaxy.
        - <name>_galaxy_parameters.dat
            A table representation of the parameters of the galaxy.
        - <name>_chain.dat
            A table representation of the Markov Chain.
            Each line holds one set of galaxy parameters and its associated reduced chi.
        - <name>_run_parameters.txt
            A plain text representation of the run_parameters.
        - <name>_instrument.txt
            A plain text representation of the instrument parameters.
        - <name>_convolved_cube.fits
            A FITS file containing the PSF-convolved result cube.
        - <name>_deconvolved_cube.fits
            A FITS file containing the pre-convolution clean cube.
        - <name>_residuals_cube.fits
            A FITS file containing the diff between input data and simulation.
        - <name>_3Dkernel.fits
            A FITS file containing the 3D kernel used

        - <name>_true_flux_map.fits
            A FITS file containing the true flux map [intrinsic]
        - <name>_true_vel_map.fits
            A FITS file containing the true velocity map [intrinsic]
        - <name>_true_sig_map.fits
            A FITS file containing the true dispersion map [intrinsic]

        - <name>_obs_flux_map.fits
            A FITS file containing the observed flux map [intrinsic]
        - <name>_obs_vel_map.fits
            A FITS file containing the observed velocity map [intrinsic]
        - <name>_obs_sig_map.fits
            A FITS file containing the observed dispersion map [intrinsic]

        - <name>_images.pdf/png
            A PNG image generated by the ``plot_images`` method.
            Note: the overwrite option is always true for this file.
        - <name>_mcmc.pdf/png
            A PNG image generated by the ``plot_mcmc`` method.
            Note: the overwrite option is always true for this file.

        - <name>_true_maps.pdf/png
            A PNG image generated by the ``plot_true_vfield`` method.
        - <name>_obs_maps.pdf/png
            The observed maps generated by the ``plot_obs_vfield`` method.

        - <name>_model.txt
            The model configuration
        - <name>_instrument.txt
            The instrument configuration

        - <name>_geweke.pdf/png
            The geweke diagnostics plot
        - <name>_galaxy_parameters_convergence.dat
            The convergence of each parameter based on the geweke diagnostics

        - <name>_corner.pdf/png
            The corner plot for the MCMC chain. Requires

        - <name>_stats.dat
            A ascii file containing the BIC/DIC etc criteria

        The .dat files can be easily read using astropy.table and its ``ascii_fixedwidth`` format : ::

            Table.read('example.chain.dat', format='ascii.fixed_width')

        .. warning::
            The generated files are not compressed and may take up a lot of disk
            space.

        name: string
            An absolute or relative name that will be used as prefix for the
            save files.
            Eg: 'my_run', or '/home/me/science/my_run'.
        overwrite: bool
            When set to true, will OVERWRITE existing files.
        """
        if self.chain is None:
            raise RuntimeError("Nothing to save! Run .run_mcmc() first.")

        filename = '%s_galaxy_parameters.txt' % name
        self._save_to_file(filename, self.galaxy.long_info(), overwrite)
        filename = '%s_galaxy_parameters.dat' % name
        self._save_to_file(filename, self.galaxy.structured_info(), overwrite)

        filename = '%s_chain.dat' % name
        self._save_to_file(filename, self._chain_as_asciitable(), overwrite)

        filename = '%s_stats.dat' % name
        self.stats.write(filename, format='ascii.fixed_width', overwrite=overwrite)

        filename = '%s_run_parameters.txt' % name
        self._save_to_file(filename, self.__str__(), overwrite)
        filename = '%s_instrument.txt' % name
        self._save_to_file(filename, self.instrument.__str__(), overwrite)
        filename = '%s_model.txt' % name
        self._save_to_file(filename, self.model.__str__(), overwrite)

        filename = '%s_convolved_cube.fits' % name
        self.convolved_cube.write_to(filename, overwrite)
        filename = '%s_deconvolved_cube.fits' % name
        self.deconvolved_cube.write_to(filename, overwrite)
        filename = '%s_residuals_cube.fits' % name
        self.residuals_cube.write_to(filename, overwrite)
        filename = '%s_3Dkernel.fits' % name
        self.psf3d.write_to(filename, overwrite)

        filename = '%s_obs_flux_map.fits' % name
        self.obs_flux_map.write_to(filename, overwrite)
        filename = '%s_obs_vel_map.fits' % name
        self.obs_velocity_map.write_to(filename, overwrite)
        filename = '%s_obs_disp_map.fits' % name
        self.obs_disp_map.write_to(filename, overwrite)

        filename = '%s_true_flux_map.fits' % name
        self.true_flux_map.write_to(filename, overwrite)
        filename = '%s_true_vel_map.fits' % name
        self.true_velocity_map.write_to(filename, overwrite)
        filename = '%s_true_disp_map.fits' % name
        self.true_disp_map.write_to(filename, overwrite)

        filename = '%s_rotcurve' % name
        self.model.plot_vprofile(self.galaxy,chain=self.sub_chain,filename=filename + '.png')
        self.model.plot_vprofile(self.galaxy,chain=self.sub_chain,filename=filename + '.pdf')

        #for quick display only
        filename = '%s_true_maps' % name
        self.plot_true_vfield(filename + '.png')
        self.plot_true_vfield(filename + '.pdf')

        filename = '%s_obs_maps' % name
        self.plot_obs_vfield(filename + '.png')
        self.plot_obs_vfield(filename + '.pdf')

        #@fixme
        # if self.error_maps:
        #    filename = '%s_true_flux_map_error.fits' % name
        #    self.true_flux_map_error.write_to(filename, overwrite)
        #    filename = '%s_true_vel_map_error.fits' % name
        #    self.true_velocity_map_error.write_to(filename, overwrite)
        #    filename = '%s_true_disp_map_error.fits' % name
        #    self.true_disp_map_error.write_to(filename, overwrite)

        filename = '%s_images' % name
        self.plot_images(filename + '.png')
        self.plot_images(filename + '.pdf')

        filename = '%s_mcmc' % name
        self.plot_mcmc(filename + '.png', method='last')
        self.plot_mcmc(filename + '.pdf', method='last')

        #Deprecicated
        # filename = '%s_correlations' % name
        #self.plot_correlations(filename + '.png')
        #self.plot_correlations(filename + '.pdf')

        try:
            filename = '%s_corner' % name
            self.corner=self.plot_corner(filename + '.png',nsigma=4)
            _ = self.plot_corner(filename + '.pdf',nsigma=4)
        except:
            self.corner=False
            self.logger.warning("plot corner failed ")

        filename = '%s_geweke' % name
        self.plot_geweke(filename + '.png')
        self.plot_geweke(filename + '.pdf')

        filename = '%s_galaxy_parameters_convergence.dat' % name
        self.convergence.write(filename, format='ascii.fixed_width', overwrite=overwrite)

        self.logger.info("Saved files in %s" % os.getcwd())

    #fixme: to do
    # def read_files(self, name):

    def __str__(self):
        """
        Return information about this run in a multiline string.
        """
        return """
galpak_version = %s

%s

mcmc_method = %s
mcmc_sampling = %s
iterations = %s
random_scale = %s

parameters method = %s, chain_fraction: %s,
CI percentile: %s,

%s

min_boundaries = %s
max_boundaries = %s
known_parameters = %s
initial_parameters = %s

final_parameters = \n %s
best_chi2 = %s
median_chi2 = %s
BIC = %s

acceptance_rate = %s
    """ % (
            self.version,
            self.instrument, self.mcmc_method, self.mcmc_sampling, self.max_iterations, self.random_scale,
            self.method,
            self.chain_fraction, self.percentile,
            self.model,
            self.min_boundaries, self.max_boundaries,
            self.known_parameters, self.initial_parameters,
            self.galaxy.structured_info(),
            self.best_chisq, self.chi_at_p,
            self.BIC,
            self.acceptance_rate
        )

    def _chain_as_asciitable(self):
        """
        Exports the chain as an `asciitable`.
        See the public API `import_chain()` for the reverse operation of loading
        the chain from an `asciitable` file.
        """
        out = StringStdOut()
        asciitable.write(self.chain,
                         output=out,
                         Writer=asciitable.FixedWidth,
                         names=self.chain.dtype.names)
        return out.content

    def _save_to_file(self, filename, contents, clobber):
        if not clobber and os.path.isfile(filename):
            raise IOError("The file '%s' already exists. Specify clobber=True to overwrite it.")
        with open(filename, 'w') as f:
            f.write(contents)

    def _read_file(self, filename):
        with open(filename, 'r') as f:
            contents=f.read()
        return contents

    def _get_min_chi_index(self):
        """
        Gets the index in the chain of the parameters with the minimal chi.
        """

        if self.chain is None:
            raise RuntimeError("No chain! Run `run_mcmc()` first.")

        idx = self.chain.idxsorted[0]

        return idx




    #############################################
    #
    # Private methods, modeling
    #
    #############################################

    def _set_verbose(self, verbose):
        """
        Update the logger's status
        """
        self.logger.disabled=False
        if verbose is True:
            #self.logger.setLevel('INFO')
            if self.model is not None:
                self.model.logger.disabled = False
            np.seterr(all='warn')
        elif verbose is False:
            #self.logger.setLevel('DEBUG')
            if self.model is not None:
                self.model.logger.disabled = True
            np.seterr(all='ignore')
        elif verbose is None:
            self.logger.disabled=True
            if self.model is not None:
                self.model.logger.disabled = True
        else:
            raise ValueError("verbose should be None | True | False")

    def _set_psf(self):
        """
        set PSF from config file
        """
        psf = None
        config = self.config['PSF']
        psf_keys={}
        for k in ['type', 'fwhm', 'alpha', 'beta', 'ba', 'pa']:
            if k in config.keys():
                psf_keys[k] = k
            elif 'psf_'+k in config.keys():
                psf_keys[k] = 'psf_'+k

        if 'type' in psf_keys:
            psf_type = config[psf_keys['type']].lower()
            self.logger.info("CONFIG: PSF: type {%s} found " % (psf_type))
        else:
            self.logger.error("CONFIG: PSF: type not specified")

        if psf_type == 'moffat':
            try:
                tmp_fwhm = float(config[psf_keys['fwhm']].split()[0])
            except:
                tmp_fwhm = None
            try:
                tmp_alpha = float(config[psf_keys['alpha']].split()[0])
            except:
                tmp_alpha = None

            if tmp_alpha is None:
                psf = MoffatPointSpreadFunction(
                    fwhm=float(config[psf_keys['fwhm']].split()[0]),  # psf_fwhm,
                    #alpha=float(self.config.get('PSF', psf_keys['alpha']).split()[0]),
                    beta=float(config[psf_keys['beta']].split()[0]),
                    pa  =float(config[psf_keys['pa']].split()[0]),
                    ba  =float(config[psf_keys['ba']].split()[0])
                )
            elif tmp_fwhm is None:
                psf = MoffatPointSpreadFunction(
                    #fwhm=self.config.get('PSF', psf_keys['fwhm']).split()[0]),  # psf_fwhm,
                    alpha=float(config[psf_keys['alpha']].split()[0]),
                    beta =float(config[psf_keys['beta']].split()[0]),
                    pa   =float(config[psf_keys['pa']].split()[0]),
                    ba   =float(config[psf_keys['ba']].split()[0])
                )
            else:
                raise self.logger.error("CONFIG: PSF: specify alpha or fwhm but not both")

        elif psf_type == 'gaussian':
            psf = GaussianPointSpreadFunction(
                    fwhm=float(config[psf_keys['fwhm']].split()[0]),  # psf_fwhm,
                    pa=float(config[psf_keys['pa']].split()[0]),
                    ba=float(config[psf_keys['ba']].split()[0])
            )
        elif psf_type == 'custom':
            if '.fits' in psf_keys['filename']:
                psf = ImagePointSpreadFunction(image_2d=config[psf_keys['filename']])
            else:
                raise self.logger.error("CONFIG: PSF: filename not supported (use .fits)")
        else:
            #@fixme add other options
            raise NotImplementedError("Currently only PSF type = `moffat` and `gaussian` supported")

        return psf

    def _set_lsf(self):
        """
        sets LSF from config LSF: lsf_fwhm
        :return:
        """
        lsf = None
        config = self.config['LSF']
        lsf_keys={}
        for k in ['type', 'fwhm']:
            if k in list(config.keys()):
                lsf_keys[k] = k
            elif 'lsf_'+k in config.keys():
                lsf_keys[k] = 'lsf_'+k

        if 'type' in lsf_keys:
            lsf_type = config[lsf_keys['type']].lower()
        else:
            lsf_type = None
            self.logger.warning("CONFIG: LSF: type not specified")

        if lsf_type == 'gaussian':
            lsf = GaussianLineSpreadFunction(
                fwhm=float(config[lsf_keys['fwhm']].split()[0])
            )  # lsf_fwhm,
        elif lsf_type is None:
            lsf = None
        else:
            #@fixme
            raise NotImplementedError

        return lsf

    def _read_instrument(self, file_config):

        if os.path.isfile(file_config):
            self.config.read(file_config)
        else:
            raise ValueError("Instrument Config file %s not present! " % (file_config))

        psf = self._set_psf()
        try:
            lsf = self._set_lsf()
        except:
            lsf = None
            self.logger.warning("CONFIG: LSF not found in config file")

        if 'INSTRUMENT' in self.config.sections():
            config = self.config['INSTRUMENT']
            myinstr = config['type'].lower()
            self.logger.info("CONFIG: INSTRUMENT: type {%s} found " % (myinstr))

            if 'pixscale' in list(config.keys()):
                scale = float(config['pixscale'].split()[0])  #
            else:
                scale = None

            if 'muse' == myinstr:
                instrument = MUSE(psf=psf, lsf=lsf)
            elif 'musewfm' == myinstr:
                instrument = MUSEWFM(psf=psf, lsf=lsf)
            elif 'musenfm' == myinstr:
                instrument = MUSENFM(psf=psf, lsf=lsf)
            elif 'alma' in myinstr:
                instrument = ALMA(psf=psf, lsf=lsf, pixscale=scale)
            elif 'sinfok250' in myinstr:
                instrument = SINFOK250(psf=psf, lsf=lsf)
            elif 'sinfok100' in myinstr:
                instrument = SINFOK100(psf=psf, lsf=lsf)
            elif 'sinfoj250' in myinstr:
                instrument = SINFOJ250(psf=psf, lsf=lsf)
            elif 'sinfoj100' in myinstr:
                instrument = SINFOJ100(psf=psf, lsf=lsf)
            elif 'harmoni' in myinstr:
                instrument = HARMONI(psf=psf, lsf=lsf, pixscale=scale)
            elif 'kmos' in myinstr:
                instrument = KMOS(psf=psf, lsf=lsf)
            elif 'osiris' in myinstr:
                instrument = OSIRIS(psf=psf, lsf=lsf)
            elif 'generic' in myinstr:
                instrument = Generic(psf=psf, lsf=lsf, default_spaxel_size=scale)
            else:
                # @fixme:generalize
                raise NotImplementedError
        else:
            #default
            self.logger.info("CONFIG: INSTRUMENT not present. Will use MUSE as default")
            instrument = MUSE(psf=psf, lsf=lsf)
        return instrument

    def _read_model(self, file_config):
        """
        sets model from config MODEL
        :return:
        """

        model = None
        if os.path.isfile(file_config):
            self.logger.info("Reading model {:s}".format(file_config))
            self.config.read(file_config)
        else:
            raise ValueError("Model Config file %s not present" % (file_config))

        if self.config.has_section('MODEL'):
            config = self.config['MODEL']
        else:
            self.logger.warning("CONFIG file has no MODEL section")

        if 'type' in list(config.keys()):
            model_type = config['type'].lower()
        else:
            self.logger.error("CONFIG: Model: type not specified")


        args={}
        #try:
        #    redshift = float(config['redshift'])
        #except:
        #    redshift = None

        #args['redshift']=redshift

        if  'default' in model_type:
            model = DefaultModel
        elif 'sersic' in model_type:
            model = ModelSersic
        else:
            raise ValueError("Model type invalid. Must be DefaultModel or ModelSersic")

        #args parameters
        import inspect
        var_args = inspect.getargspec(model).args
        for k in var_args[1:]:
            if k in list(config.keys()):
                try:
                    args[k]=float(config[k])
                except:
                    args[k]=config[k]


        return model(**args)

    def _read_params(self, file_config, type):
        """
        sets random scale from config file
        :return: ModelParamers
        """
        if os.path.isfile(file_config):
            self.config.read(file_config)
        else:
            raise ValueError("Random Scale: Config file %s not present" % (file_config))

        if self.config.has_section(type):
            config = self.config[type]
        else:
            self.logger.warning("Config file has no %s section" % (type))

        par = self.model.parameters_class()()

        for r in list(config.keys()):
            if r in par.names:
                par[r] = float(config[r])
            else:
                self.logger.warning("Config %s has keys not used for this model" (type))
        return par

    def _init_sampling_scale(self, random_scale, should_guess_flags):
        dim_d = np.size(self.cube.data)
        dim_p = len(self.initial_parameters)

        # Tweak the random amplitude vector (Kp coeff, as pid)
        # that we can document Model.setup_random_amplitude() adequately
        random_amplitude = np.sqrt(
            (self.min_boundaries - self.max_boundaries) ** 2 / 12.
        ) * dim_p / dim_d

        # Let the model adjust the random amplitude of the parameter jump
        self.model.setup_random_amplitude(random_amplitude)

        # Scale MCMC if needed // allowing vectors
        if random_scale is not None:
            if np.size(random_scale) != 1:
                merge_where_nan(random_scale, np.ones_like(random_amplitude))
            random_amplitude = random_amplitude * random_scale

        # Zero random amplitude where parameters are known
        random_amplitude = random_amplitude * should_guess_flags

        return random_amplitude



#!/usr/bin/env python
# coding=utf-8

# DISCLAIMER: I am not sure this will work properly for everyone.
#             Please report bugs.

# Run this script to leverage the power of `distutils` to install or build.

# Doc :
# - https://docs.python.org/2/install/index.html
# - https://docs.python.org/2/distutils/setupscript.html

import io
from os import path
from setuptools import setup, find_packages

# Read version.py
__version__ = None
with io.open('VERSION') as f:
    exec(f.read())

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='galpak',
    version=__version__,
    author='Nicolas Bouché',
    author_email='nicolas.bouche@univ-lyon1.fr',
    long_description = long_description,
    long_description_content_type='text/markdown',
    url='http://galpak3d.univ-lyon1.fr',
    license='GPL',
    description='A tool to extract the intrinsic (i.e. deconvolved) Galaxy '
                'Parameters and Kinematics from any 3-Dimensional data.',
    packages=find_packages('src', ),
    package_dir ={'': 'src'},
    python_requires='>=3.5',
    install_requires=['astropy>=2.0', 'numpy>=1.14', 'scipy', 'matplotlib>=2.0'],
)



# coding=utf-8


class GalaxyParameter(property):
    """
    A parameter of a galaxy model.
    This is deep voodoo. The base class of this (property) is a python trick to allow us to define
    custom accessors (fget) and mutators (fset) for each property created with it.
    We simply use it to fetch the corresponding value in the ndarray, by numerical index.
    This allows us named/documented/autocompleted access to the galaxy parameters, instead of
    accessing them only through the numerical index, numpy-like, which is boring/annoying to memorize.
    You still can access the data through numerical indices if you want.
    Win-win ! Except for the maintenance of the particular piece of code.

    - name (string) : a name, for access (MUST be the same as the name in `GalaxyParameters.names`.)
    - key (int) : the position in the the ndarray of this parameter
    - short (string) : a shorter name, for compact display
    - doc (string) : some documentation that will appear to the enduser.
    - unit (string) : the unit, if any
    - precision (string) : the precision to use during string casting, as used by `string.format()`.
    """
    def __init__(self, name, key, short=None, doc=None, unit=None, format='3.2f'):
        self.name = name
        self.key = key
        self.unit = unit
        self.format = format
        if unit is not None:
            doc += " (%s)" % unit
        if short is not None:
            self.short = short
        else:
            self.short = name

        super(GalaxyParameter, self).__init__(fget=lambda selfie: selfie[key],
                                              fset=lambda selfie, value: selfie.__setitem__(key, value),
                                              doc=doc)
        # The above doc= is insufficient to set the docstring?! This takes care of it.
        self.__doc__ = doc


# -*- coding: utf-8 -*-

import numpy as np
from scipy.interpolate import interp1d
from scipy import ndimage
from copy import deepcopy

try:
    import bottleneck as bn
except ImportError:
    import numpy as bn


from .hyperspectral_cube import HyperspectralCube as HyperCube
from .galaxy_parameters import  GalaxyParameters

# CONSTANTS
G = 6.67384e-11  # Gravitational Constant (m^3.kg^-1.s^-2)
SOL_MASS = 1.98e30  # Solar Mass (kg)
PARSEC = 3e16    # Parsec (m)

# LOGGING CONFIGURATION
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('GalPaK: DiskModel ')

#Python3 compatibility
try:
  basestring
except NameError:
  basestring = str

class Model:
    """
    An **interface** for your simulation models.
    The `runner` parameter is the instance of `galpak.Runner` using this model.
    """

    logger = logger
    flux_map_user = None

    def set_cosmology(self, cosmo_string='planck15'):
        try:
            from colossus.cosmology import cosmology
            from colossus import halo
            from colossus.halo import mass_defs, mass_adv, mass_so, concentration
            self.logger.info("setting cosmology with colossus package")
            if cosmo_string is cosmology.cosmologies.keys():
                cosmo=cosmology.setCosmology(cosmo_string)
            else:
                self.logger.info("colossus: cosmology %s is not available\n "
                                 "Please use one of %s \n"
                                 "Will be uing cosmology Planck15 as default" \
                                 % (cosmo_string, cosmology.cosmologies.keys()) )
                cosmo=cosmology.setCosmology('planck15')
            #put h in
            self.h = cosmo.h
            self.Ez = cosmo.Ez(self.redshift)
            DA = cosmo.angularDiameterDistance(self.redshift) / self.h  ####!!
            self.DeltaVir = halo.mass_so.deltaVir(self.redshift)
            self.halo = halo #Diemer Halo module for NFW halo models
        except ImportError:
            try:
                from astLib import astCalc
                #using  astLib for cosmology
                self.logger.info("Using AstLib for a 737 cosmology, h,OM, OL= (0.7, 0.3, 0.7)")
                self.h = astCalc.H0 / 100.
                self.Ez = astCalc.Ez(self.redshift)
                DA = astCalc.da(self.redshift)#for h=0.7
                self.DeltaVir = astCalc.DeltaVz(self.redshift)
                self.halo = None
                #self.logger.warning('Best to use colossus package from Diemer& Kratsov for this module')
            except ImportError:
                #using astropy
                from astropy import cosmology
                if cosmo_string in cosmology.parameters.available:
                    cosmo = cosmology.default_cosmology.get_cosmology_from_string(cosmo_string)
                else:
                    self.logger.info("Astropy: Parameter %s is not available\n "
                                     "Please use one of %s \n"
                                     "Will be uing astropy.cosmology Planck15 as default" \
                                     % (cosmo_string, cosmology.parameters.available))
                    cosmo = cosmology.default_cosmology.get_cosmology_from_string('Planck15')
                self.h = cosmo.h
                DA = cosmo.angular_diameter_distance(self.redshift).value #in Mpc
                self.Ez = cosmo.H(self.redshift).value/cosmo.H0
                x = cosmo.Om(self.redshift) - 1.0 #Omega_M(z) - 1.0
                self.DeltaVir = 18 * np.pi**2 + 82 *x - 39 * x**2
                self.halo = None
                #self.logger.warning('Best to use colossus package from Diemer& Kratsov for this module')

        self.kpc = DA * 1e3 * np.radians(1./3600)# in unit kpc
        #self.hkpc = self.kpc * h

    def Parameters(self):
        """
        Returns `GalaxyParameters' instance
        :return:
        """
        return GalaxyParameters()

    def parameters_class(self):
        """
        Returns the class of parameters extending `ModelParameters` that this
        model uses. Note that this returns the class itself, not an instance.
        """
        return GalaxyParameters

    def initial_parameters(self, runner):
        """
        Returns an instance of a class extending `ModelParameters`.
        You may omit parameters, they will be automatically set to the mean of
        the max and min.
        """
        raise NotImplementedError()

    def min_boundaries(self, runner):
        """
        Returns an instance of a class extending `ModelParameters`.
        You MUST provide all the parameters.
        """
        raise NotImplementedError()

    def max_boundaries(self, runner):
        """
        Returns an instance of a class extending `ModelParameters`.
        You MUST provide all the parameters.
        """
        raise NotImplementedError()

    def sanitize_parameters(self, parameters):
        """
        Mutates the parameters with custom logic.
        This is called right after parameter jumping in the loop.
        """
        #    # Handle reflection in parameters
        #    if parameters.pa > 90.:
        #        parameters.pa -= 180.
        #        parameters.maximum_velocity *= -1
        #    if parameters.pa < -90.:
        #        parameters.pa += 180.
        #        parameters.maximum_velocity *= -1

        inc = parameters['inclination']
        if inc<0:
            parameters.pa +=180

        parameters.pa = parameters.pa % 360. #to deal with circularity


    def sanitize_chain(self, chain):
        """
        Mutates the `chain` with custom logic.
        This is called at the end of the loop, before storing the chain.
        """
        pa = chain['pa']
        pa_corr = pa % 360
        chain['pa'] = pa_corr

    def setup_random_amplitude(self, amplitude):
        """
        Mutates the random `amplitude` of the parameter jump with custom logic.
        This is called during the setup of the MCMC runner.
        The amplitude is already filled with :
            sqrt((min_boundaries - max_boundaries) ** 2 / 12.) * p / v
        where
            p = number of parameters in the model
            v = number of voxels in the cube
        """
        raise NotImplementedError()

    #CUSTOM METHODS
    def set_flux_profile(self, galaxy, radius_cube):
        """
        creates a flux_cube in 3D (x,y,z)
        returns flux_cube
        """
        raise NotImplementedError()

    def set_velocity_profile(self, galaxy, radius_cube):
        """
        creates a velocity profile for circular orbits
        return v_profile
        """
        raise NotImplementedError()

    # DEFAULT METHODS

    def set_thickness_profile(self, nz, hz, flux_cube):
        """
        add the thickness profile to the flux cube
        returns flux_cube
        """
        if self.thickness_profile == 'gaussian':
            tmp_cube = np.exp(-nz**2. / 2. / hz**2.)
        elif self.thickness_profile == 'exponential':
            tmp_cube = np.exp(-np.abs(nz) / hz)  # exponential z-profile
        elif self.thickness_profile == 'sech2':
            tmp_cube = np.cosh(-nz / hz)**(-2.)  # sech^2 z-profile
        elif self.thickness_profile == 'none':
            self.logger.warning("Using no disk thickness profile: "
                                 "this galaxy will become a cylinder!")
            tmp_cube = np.ones_like(flux_cube)
        else:
            raise NotImplementedError("Context profile not supported. Should be '%s'" %
                                      self.THICKNESS_VALID)
        #normalize gaussians
        #norm = tmp_cube.sum(0)

        return flux_cube * tmp_cube


    def set_dispersion_profile(self, hz, vtot, radius_cube):
        """
        set disk dispersion profile
        """
        ### ISOTROPIC but this IS the 1-d dispersion in z-direction
        # A) using R/h=0.5 (V/sigma)^2 for very thin disk
        if self.dispersion_profile == 'thin':
            # dimension of r [3D-spatial]
            sig_ale_disk = np.sqrt(hz * vtot ** 2. / radius_cube)
        # B) using R/h=V/Sigma for compact disk
        elif self.dispersion_profile == 'thick':
            # dimension of r [3D-spatial]
            sig_ale_disk = hz * vtot / radius_cube
        # C) Using Binney Merrifield
        # elif self.dispersion_profile == 'infinitely_thin':
            # dimension of r [3D-spatial]
            # sig_ale_disk = np.sqrt(math.tau * Surf_density * G * hz_kpc)
        else:
            raise ValueError("Disk dispersion is not valid. Should be '%s' " %
                             self.DISPERSION_VALID)

        return sig_ale_disk


    ###########################################
    #HEART of model
    ###########################################
    def _create_maps(self, galaxy, shape):
        """
        Creates and returns:
        - a cube sized like (`shape`) containing a clean simulation
          of a galaxy for the `parameters`.
        - the velocity map, which is an image.
        - the dispersion (sigma) map, which is an image.

        This is called on each loop, to compute the likelihood of each generated
        set of parameters. Therefore, it needs to be fast.

        parameters: DiskParameters
            The parameters upon which the simulated disk galaxy will be built.
        shape: Tuple of 3
            The 3D shape of the resulting cube.
            Eg: (21,21,21)

        Returns a tuple.
        """
        # Create radial velocities and radial dispersions

        flux_cube, vz, vz_map, s_map, sig_map, sigz_disk_map, sig_intr = \
            self._compute_galaxy_model(galaxy, shape)

        # normalize to real flux
        flux_map = flux_cube.sum(0)
        flux_map = galaxy.flux * flux_map / flux_map.sum()

        return flux_map, vz_map, s_map

    def _create_cube(self, galaxy, shape, z_step_kms, zo=None):
        """
        Creates and returns:
        - the flux map, of shape (shape[1], shape[0]) of a galaxy
        - the velocity map, which is an image.
        - the dispersion (sigma) map, which is an image.

        This is called on each loop, to compute the likelihood of each generated
        set of parameters. Therefore, it needs to be fast.

        parameters: DiskParameters
            The parameters upon which the simulated disk galaxy will be built.
        shape: Tuple of 3
            The 3D shape of the resulting cube.
            Eg: (21,21,21)

        Returns a tuple.
        """
        # Create radial velocities and radial dispersions

        flux_cube, vz, v_map, s_map, sig_map, sigz_disk_map, sig_intr = \
            self._compute_galaxy_model(galaxy, shape)

        # normalize to real flux
        flux_map = flux_cube.sum(0)
        flux_map = galaxy.flux * flux_map / flux_map.sum()

    #def _create_cube(self, shape, f_map, v_map, s_map, z_step_kms, zo=None):
    #    """
    #    Creates a 3D cube from Flux, Velocity and Dispersion maps.

    #    shape: tuple
    #        Defines the created cube's shape
    #    f_map: ndarray
    #        2D image of the flux
    #    v_map: ndarray
    #        2D image of the velocity
    #    s_map: ndarray
    #        2D image of the dispersion
    #    z_step_kms: float
    #        in km/s
    #    zo: int
    #        The centroid (in pixel) in the z-direction
    #        If not set, will be the middle of the z-axis
    #    """

        vmax = z_step_kms * (shape[0] - 1) / 2.

        if zo is None:
            vzero = vmax
        else:
            vzero = z_step_kms * zo  # fixme: shouldn't this be with zo-0.5 ?

        # Array of velocities : array of indices from to 0 to Vmax
        velocities = np.linspace(0, 2 * vmax, shape[0]) - vzero

        # Create a copy of Vmap in each z plane
        tmpcube = np.outer(np.ones_like(velocities), v_map)
        vcube = tmpcube.reshape(shape)

        # Create a copy of Smap in each z plane
        tmpcube = np.outer(np.ones_like(velocities), s_map)
        scube = tmpcube.reshape(shape)

        # Create a copy of velocities in each location
        if z_step_kms > 0:
            vbegin = np.min(velocities)
            vfinal = np.max(velocities)
        else:
            vbegin = np.max(velocities)
            vfinal = np.min(velocities)
        # Create grid of velocities
        vgrid, y, x = np.mgrid[
            vbegin: vfinal + 0.9 * z_step_kms: z_step_kms,
            0: np.shape(v_map)[0],
            0: np.shape(v_map)[1]
        ]

        # Apply Gaussian
        finalcube = np.exp(-0.5 * (vgrid - vcube) ** 2 / (scube ** 2))

        # For doublets, add the blue component
        # deltaV is defined as (l2-l1)/l2 * c ie from the reddest line
        if self.line is not None:
            # Delta between emission peaks (in km/s ?)
            # line['wave']=[l, lref] eg. [3726.2, 3728.9] # Observed or Rest;
            # line['ratio']=[0.8, 1.0] eg. [0.8, 1.0] ; the primary line for redshifts is the reddest
            delta = 3e5 * (self.line['wave'][1]-self.line['wave'][0])/(self.line['wave'][0]+self.line['wave'][1])*2
            ratio = self.line['ratio'][0] / self.line['ratio'][1]
            #ratio = galaxy.line_ratio

            # Add the blue doublet
            finalcube = finalcube + ratio * np.exp(-0.5 * (vgrid - (vcube - delta)) ** 2 / (scube ** 2))

        # delta = 217.  # km/s OII
        # delta = -945  # km/s Ha 6564.614 NII 6585.27
        # delta = 2872  # km/s OIII 4960.295 5008.240
        # delta = 641   # km/s SII 6718.29  6732.67

        # Normalize from amplitude ratios int f dv = int f dz
        ampsum = bn.nansum(finalcube,0)
        #amplitude = f_map / bn.nansum(finalcube, 0)
        #amplitude[np.isnan(amplitude)] = 0
        amplitude =  flux_map / ampsum
        amplitude[np.isfinite(amplitude)==False] = 0
        #amplitude =  np.where(np.isfinite(amplitude),f_map / ampsum, 0)

        # Create a cube of Amplitude in each z plane
        amplitude_cube = np.resize(amplitude, shape)

        finalcube = finalcube * amplitude_cube
        finalcube[np.isfinite(finalcube)==False] = 0

        cube = HyperCube(data=finalcube)

        return cube, flux_map, v_map, s_map

    def _compute_galaxy_model(self, parameters, shape):
        """
        Create the disk galaxy model related cubes and maps.

        parameters: DiskParameters
            The parameters of the disk galaxy.

        Returns
            3D flux cube,
            3D Vz cube,
            2D Vmap (1st moment of Vz),
            2D S_map,
            2D (2nd moment of Vz),
            2D disk_dispersion,
            2D intrinsic
        """
        if 'radius' in parameters.names:
            hz = parameters.radius * self.q ##0.15
        elif 'disk_radius' in parameters.names:
            hz = parameters.disk_radius * self.q ##0.15
        else:
            raise Exception("Parameter must have a radius")
        # trick to go faster, as we know the max z shape the resulting model will fit in
        #       z_shape = np.ceil(shape[1] / 2.5 * 0.15 * 6.) ## 6 sigmas
        #       z_shape = np.ceil(self.max_boundaries.radius / 2.5 * 0.15 * 6.) ## 6 sigmas
        #       or
        #       z_shape = np.ceil(self.max_boundaries.radius * 2 * 3 * 0.15) # 3 hz_max on each side
        # this short cut introduced a bug in the flux_deconvolved maps! fixed after 1.6.0
        z_shape = int(np.ceil ( (shape[1] + shape[2]) / 2.)) #mean of spatial dimensions
        self.logger.debug(" using z_shape %d in compute_galaxy_model" % (z_shape) )
        self.logger.debug(" cube is %d in z- " % (shape[0]) )
        # Minimum shape for z direction -- collapsed, so big sizes unimportant
        # z_shape = 6 hz(sigma)_max;
        # hz_max = size_box/2.5 (proxy for radius max) * 0.15

        # Create arrays of indices in image frame; this cube has spatial dimentions in x,y,z; so should be almost a perfect cube in shape.
        ind_shape = tuple([z_shape, shape[1], shape[2]])
        z, y, x = np.indices(ind_shape, dtype='float64')

        #Spatial Cubes and Vz from spatial dimensions ind_shape
        nx, ny, nz, vz_cube, vtot = self._map_indices(x, y, z, parameters)

        radius_cube = np.hypot(nx,ny) #np.sqrt(nx**2. + ny**2.)

        flux_cube = self.set_flux_profile(parameters, radius_cube)

        # Adjust disk thickness
        flux_cube = self.set_thickness_profile(nz, hz, flux_cube)

        # Normalize by flux
        total = bn.nansum(flux_cube) #should always be >0
        if total>0:
            flux_cube =  flux_cube / total
        else:
            raise ValueError("Something is wrong. Total flux is zero")

        # Flux weighted
        flux_map = bn.nansum(flux_cube, 0)
        bad = (flux_map**2 == 0)

        vz_map = bn.nansum(vz_cube * flux_cube / flux_map, 0)
        vz_map[bad] = np.nan  #force outer to 0

        v2_map = bn.nansum(vz_cube ** 2 * flux_cube / flux_map, 0)
        Var_map = v2_map - vz_map ** 2
        sig_map_disk = np.where(Var_map>0, np.sqrt(Var_map), 0)



        ### Add Sigma_Aleatoire_disk
        sig_ale_disk = self.set_dispersion_profile(hz, vtot, radius_cube)
        # Compute weighted mean for Sig_disk
        # a) 2D map = flux-weighted map
        #Sigz_disk_map = bn.nansum(Sig_disk*A,0)/Flux_map
        # b) 2D map = flux-weighted map rms
        #sigz_map_ale = np.sqrt(bn.nansum(sig_ale_disk**2*A,0)/Flux_map)
        # c) Normalized with flux squared
        norm =  bn.nansum(flux_cube * flux_cube, 0)
        sigz_map_ale = np.sqrt(bn.nansum(sig_ale_disk ** 2 * flux_cube ** 2, 0) / norm)


        # Intrinsic Dispersionmodel
        if 'velocity_dispersion' in parameters.names:
            sig_intr = np.ones_like(sigz_map_ale) * parameters.velocity_dispersion
        else:
            sig_intr = np.ones_like(sigz_map_ale)

        #correct outer regions
        #sigz_map_ale = np.where(bad, sig_intr, sigz_map_ale)

        # Total S-map
        s_map = np.sqrt(sig_map_disk ** 2 + sigz_map_ale ** 2 + sig_intr ** 2)
        s_map[bad]=np.nan

        return flux_cube, vz_cube, vz_map, s_map, \
            sig_map_disk, sigz_map_ale, sig_intr

    def _map_indices(self, xx, yy, zz, parameters):
        """
        Takes arrays of indices, galaxy parameters, and rotates accordingly.
        Returns xx, yy, zz indices of object in image frame
        Returns z-component of velocity of object in image frame
        """

        inclination = np.radians(parameters.inclination)
        pa = np.radians(parameters.pa - 90)

        zo = (np.size(zz, 0) - 1 ) / 2.  # center along z

        # Rotate inclination (around x)
        rot_i = self._make_rotation_matrix_around_x(inclination)
        rot_mi = self._make_rotation_matrix_around_x(-inclination)

        # Rotate PA (around z) anti-clockwise
        rot_pa = self._make_rotation_matrix_around_z(pa)
        rot_mpa = self._make_rotation_matrix_around_z(-pa)

        # Transformation from sky coords to disk plane
        xx = xx - parameters.x
        yy = yy - parameters.y
        zz = zz - zo

        # Rotation around z for PA
        nx = rot_pa[0, 0] * xx + rot_pa[0, 1] * yy
        ny = rot_pa[1, 0] * xx + rot_pa[1, 1] * yy
        nz = rot_pa[2, 2] * zz

        # Rotation around x for inclination
        x = rot_i[0, 0] * nx
        y = rot_i[1, 1] * ny + rot_i[1, 2] * nz
        z = rot_i[2, 1] * ny + rot_i[2, 2] * nz

        # Compute the radius cube
        radius_cube = np.sqrt(x ** 2. + y ** 2.)

        # Circular orbits -- we're DIVIDING BY ZERO ; it's okay, need for speed
        vx = np.where(np.isfinite(y / radius_cube),  +y / radius_cube, 0)
        vy = np.where(np.isfinite(x / radius_cube),  -x / radius_cube, 0)

        if 'virial_velocity' in parameters.names:
            radius_3d = np.sqrt(x**2. + y**2. + z**2.)
            v_profile = self.set_velocity_profile(parameters, radius_cube, radius_3d)
        else:
            v_profile = self.set_velocity_profile(parameters, radius_cube)
        # Remove NaNs where radius is 0
        v_profile[radius_cube == 0] = 0.

        # 3D velocity vectors
        vx = vx * v_profile
        vy = vy * v_profile

        # Rotation of [vx, vy, 0] -inclination around x
        vvx = rot_mi[0, 0] * vx
        vvy = rot_mi[1, 1] * vy
        vvz = rot_mi[2, 1] * vy

        # Rotation of -pa around z
        vx = rot_mpa[0, 0] * vvx + rot_mpa[0, 1] * vvy
        vy = rot_mpa[1, 0] * vvx + rot_mpa[1, 1] * vvy
        vz = rot_mpa[2, 2] * vvz

        v = np.sqrt(vx ** 2 + vy ** 2 + vz ** 2)

        return x, y, z, vz, v

    def __str__(self):
        dic=self.__mdict__()
        model = """[MODEL] :
  type = {i}""".format(i=self.__name__())
        for k in dic.keys():
            model += """
  {k} = {v} """.format(k=k,v=dic[k])

        return model

    def __name__(self):
        return self.__class__.__name__

    def __mdict__(self):
        dic = vars(self).copy() #@fixme should deepcopy
        dic['model'] = self.__name__()
        k = dic.pop('flux_map_user') if 'flux_map_user' in dic.keys() else None

        return dic

    ################### PRIVATE methods
    ####################################

    @staticmethod
    def _make_rotation_matrix_around_x(a):
        return np.array([[1,         0,          0],
                         [0, np.cos(a), -np.sin(a)],
                         [0, np.sin(a),  np.cos(a)]])

    @staticmethod
    def _make_rotation_matrix_around_z(a):
        return np.array([[np.cos(a),  np.sin(a), 0],
                         [-np.sin(a), np.cos(a), 0],
                         [0,          0,         1]])


    @staticmethod
    def _int_x_dev(x):
        """
        (-4 (5040 + 5040 x^(1/4) + 2520 sqrt(x) + 840 x^(3/4) + 210 x + 42 x^(5/4) + 7 x^(3/2) + x^(7/4)))/exp(x^(1/4))
        """
        return (-4. * (5040. + 5040. * x ** 0.25 + 2520. * np.sqrt(x) + 840. * x ** 0.75 + 210. * x + 42. * x ** 1.25 + 7. * x ** 1.5 + x ** 1.75)) * np.exp(-x ** 0.25)

    ##################################
    def _user_profile(self, parameters, radius_1d, resample_by=10):
        """
            returns 1D profile profile from user map
        :param parameters:
        :param radius_cube:
        :return:
        """
        shape = self.flux_map_user.shape #shape of image2d

        # Create arrays of indices in image frame; this cube has spatial dimentions in x,y,z; so should be almost a perfect cube in shape.
        ind_shape = tuple([21, shape[0]*resample_by, shape[1]*resample_by])#oversampling x10
        z, y, x = np.indices(ind_shape)

        galaxy = deepcopy(parameters)
        galaxy.x = galaxy.x * resample_by
        galaxy.y = galaxy.y * resample_by
        # Spatial Cubes and Vz from spatial dimensions ind_shape
        nx, ny, nz, vz_cube, vtot = self._map_indices(x, y, z, galaxy)

        radius_2d  = np.hypot(nx, ny)  # np.sqrt(nx**2. + ny**2.)
        rad = radius_2d[10,:,:]

        ##PA is from x axis clock-wise  // or y anti-clockwise
        #y,x = np.indices(self.flux_map_user.shape)
        #ang=np.radians(parameters['pa'])
        #ba = np.cos(np.radians(parameters['inclination'])) #approx
        #dx=parameters['x']-x
        #dy=parameters['y']-y
        #dx_p=dx*np.cos(ang)-dy*np.sin(ang)
        #dy_p=dx*np.sin(ang)+dy*np.cos(ang)
        ##rotation matrix
        #rad = np.sqrt( dx_p**2+dy_p**2/ba**2  )

        #flux_cube = self.set_flux_profile(parameters, radius )

        # xp=np.arange(rmax)
        xp = np.r_[0:radius_1d.max()+1.:1./resample_by]*resample_by  # starts at 1,2,3,4 pixels
        indata = self.flux_map_user.copy() #2d map
        indata = ndimage.zoom(indata, (resample_by, resample_by))

        mysum = np.array([indata[rad <= p].sum() for p in xp])

        # Compute SB profile
        # area=pi*(np.arange(rmax))**2*ba_raw
        dx = xp[1]-xp[0]
        on = np.ones(indata.shape) * dx

        area = [on[rad <= p].sum() for p in xp]
        diff_area = np.where(np.diff(area) == 0, dx, np.diff(area))

        SBprofile = np.diff(mysum) / diff_area #/ (pixscale * pixscale)
        SBprofile = np.hstack([SBprofile[0],SBprofile]) #to avoid infty at zero

        f = interp1d(xp,SBprofile, kind='cubic')
        SBprofile_interp = f(radius_1d*resample_by)

        return SBprofile_interp

    def compute_Mdyn_at_Rhalf(self, parameters, radius=None):
        """
        From parameters and pixscale computes Mdyn(<Rhalf)

        """
        if radius is None:
            radius = parameters['radius']

        if self.pixscale is None:
            raise ValueError("Pixel scale undefined")
        else:
            pixscale = self.pixscale

        # Compute angular distance in Mpc
        # Convert it to kpc/arcsec
        rhalf_kpc = radius * pixscale * self.kpc  # in kpc

        #define velocity function
        v_rhalf = self.set_velocity_profile(parameters, radius)

        # v² = (GM/r) in km/s
        dynamic_mass = self._mass_from_GMr(v_rhalf, rhalf_kpc)#computed at R1/2.

        return dynamic_mass

    def compute_MvirRvir(self, Vvir):
        """
        computes Mvir, Rvir from Vvir
            uses either colossus R_to_M or GM/R
        :param Vvir:
        :return:
        """

        ##Rvir in kpc/h
        #Rvir = Vvir /  np.sqrt(self.DeltaVir/2.) / (100 * self.Ez/1e3) #in kpc

        #Rvir in kpc
        Rvir = Vvir /  np.sqrt(self.DeltaVir/2.) / (100 * self.h * self.Ez) * 1e3 #in kpc

        Mvir = self._mass_from_GMr(Vvir, Rvir)

        return Mvir, Rvir

    def _colossus_Mvir_compare(self, Vvir):

        Rvir = Vvir /  np.sqrt(self.DeltaVir/2.) / (100 * self.h * self.Ez) * 1e3 #in kpc

        ###############shape independent /cosmology dependent
        mdef='vir'
        Mvir = self.halo.mass_so.R_to_M(R=Rvir*self.h,z=self.redshift,mdef='vir')#kpc

        return Mvir, self._mass_from_GMr(Vvir, Rvir * self.h)

    @staticmethod
    def _mass_from_GMr(velocity, radius):
        """
        compute mass from GM/r=v2
        :param velocity: in km/s
        :param radius: in kpc
        :return:
        """
        return (1e3 * velocity) ** 2 * (PARSEC * radius * 1e3) / G / SOL_MASS

    def _fv_newton(self, r, radius, vmax):
        """
        Returns the velocity for the mass profile given the r-cube.
        The cumulative flux I(<r) profile is analytically calculated
        Only for n=1,0.5,4.
        """

        mass = 1  # enclosed_mass for Gaussian

        if self.flux_profile == 'exponential':
            rr = 1.68 * r / radius
            mtot = mass * (1. - rr * np.exp(-rr) - np.exp(-rr))
        elif self.flux_profile == 'gaussian':
            mtot = mass * (1. - np.exp(-r ** 2 / 2. / (2 * radius / 2.35) ** 2))
        elif self.flux_profile == 'de_vaucouleurs':
            mtot = mass * (self._int_x_dev(r / radius) - self._int_x_dev(0))
        else:
            raise ValueError("Flux profile is not valid. Only for sersic_n=[0.5, 1, 4].")

        v = np.sqrt(mtot / (r + 1e-9))  # in km/s

        # Normalization set by Vmax (sets the mass)
        if np.size(v) > 1:
            v = v / bn.nanmax(v) * vmax
        else:
            v = v / np.max(v) * vmax

        return v




# -*- coding: utf-8 -*-
from __future__ import division

import logging
import math


import numpy as np

from astropy.io import fits

logging.basicConfig(level=logging.INFO)
logHC = logging.getLogger('GalPak: HyperCube')


# Some todos
#
#Sanitization operations :
#    - IDL-made FITS :
#        - keyword FSCALE (http://astro.uni-tuebingen.de/software/idl/astrolib/fits/mrdfits.html)
#        - other non-ISO header blunders -_-
#    - CDEL_3 => CDELT3
#
#Use SI units, http://www.aanda.org/articles/aa/full_html/2010/16/aa15362-10/aa15362-10.html#S39
#Uniformization operations :
#    - Spatial Units !
#    - Spectral Units !

class HyperspectralCube:
    """
    A hyperspectral cube has 3 dimensions : 2 spatial, and 1 spectral.
    It is basically a rectangular region of the sky seen on a discrete subset of
    the light spectrum. It is heavily used in spectral imaging, and is provided
    by spectroscopic explorers such as MUSE.

    This class is essentially a convenience wrapper for data values and header
    specs of a single hyperspectral cube.

    It understands the basic arithmetic operations ``+ - * / **``,
    which should behave much like with numpy's ndarrays, and update the header accordingly.
    An operation between a cube and a number will behave as expected, by applying the operation voxel-by-voxel
    to the cube. (it broadcasts the number to the shape of the cube)
    An operation between a cube and an image (a slice along the spectral axis, z) will behave as expected,
    by applying the operation image-by-image to the cube. (it broadcasts the image to the shape of the cube)

    It also understands indexation of the form ``cube[zmin:zmax, ymin:ymax, xmin:xmax]``.

    See also http://en.wikipedia.org/wiki/Hyperspectral_imaging .

    Note : no FSCALE management is implemented yet.


    Factories
    ---------

    You can load from a single HDU of a FITS (Flexible Image Transport System) file : ::

        HyperspectralCube.from_file(filename, hdu_index=None, verbose=False)


    Parameters
    ----------

    data: numpy.ndarray|None
        3D numpy ndarray holding the raw data.
        The indices are the pixel coordinates in the cube.
        Note that in order to be consistent with astropy.io.fits,
        data is indexed in that order : λ, y, x.

    header: fits.Header|None
        http://docs.astropy.org/en/latest/io/fits/api/headers.html
        Note: this might become its own wrapper class in the future.

    verbose: boolean
        Set to True to log everything and anything.

    """
    xy_step = None
    z_cunit = 'Undef'
    z_step = None
    z_central = None

    def __init__(self, data=None, header=None, verbose=False, filename=None, variance=None):
        self.data = data
        self.header = header
        self.verbose = verbose
        self.filename = filename
        self.var = variance

        if header is not None:
            if not isinstance(header, fits.Header):
                raise TypeError("Header must be a fits Header")
            # Sanitize
            self.sanitize()
            #initialize
            try:
                self.initialize_self_cube()
            except:
                pass

    def defaults_from_instrument(self, instrument=None):
        """
        First reads the header for
            xy_step (CDELT1 or CD1_1)
            z_step (CDELT3 or CD3_3)
            z_cunit (CUNIT3)
            crpix (CRPIX3)
            crval (CRVAL3).
        If some of these are None, get the default values from the instrument
        and then reset the header :
            xy_step: Instrument.xy_step --> self.xy_step
            z_step:  Instrument.z_step -->  self.z_step --> CDELT3
            z_cunit: Instrument.z_cunit --> self.z_cunit --> CUNIT3
            crpix:  shape[0]/2 --> CRPIX3
            crval: Instrument.z_central --> self.z_central --> CRVAL3

         for CUNIT3 CRVAL3 CRPIX3 CDELT3
        """

        xy_step = self.get_steps()[1]
        z_step = self.header.get('CDELT3')
        z_cunit = self.header.get('CUNIT3')
        crpix = self.header.get('CRPIX3')
        crval = self.header.get('CRVAL3')

        if xy_step is None:
            self.xy_step = instrument.cube_default_xy_step

        if z_step is None:
            self.z_step = instrument.cube_default_z_step
            self.header.set('CDELT3', self.z_step)
            if 'CD3_3' in self.header.keys():
                self.header.remove('CD3_3')

        if z_cunit is None and instrument.cube_default_cunit is not None:
            self.z_cunit = instrument.cube_default_cunit
            self.header.set('CUNIT3', self.z_cunit)
        if crval is None and instrument.cube_default_z_central is not None:
            self.z_central = instrument.cube_default_z_central
            self.header.set('CRVAL3', self.z_central)
        if crpix is None:
            self.header.set('CRPIX3', (self.shape[0]) / 2.)

    def initialize_self_cube(self):
        """
        Initialize steps xy_steps z_steps z_cunnit and z_central
        z_central requires CRPIX3 CRVAL3 CDELT3
        """

        steps = self.get_steps()
        if (steps[1] is not None) and (steps[2] is not None):
            if np.abs(np.abs(steps[1]) - np.abs(steps[2])) > 1e-2:
                # when the size of pixels differ by more than 1%
                raise NotImplementedError(
                    "GalPak does not support unsquare pixels")
            xy_step = steps[1]
            if (self.header.get('CUNIT1') is not None):
                if ('deg' in self.header.get('CUNIT1') ):
                    xy_step *= 3600  # convert from deg to "
                if ('mas' in self.header.get('CUNIT1') ):
                    xy_step /= 1e3 # convert from mas to "
            else:
                raise ValueError("The Cube has no cunit1 for the spaxels, "
                                 "use cunit1 to provide it")

            self.xy_step = xy_step

        self.z_step = steps[0]

        if self.header.get('CUNIT3') is not None:
            self.z_cunit = self.header['CUNIT3']
        else:
           raise ValueError("The Cube has no cunit3 for the spaxels, "
                                 "use cunit3 to provide it")

        if self.z_central is None:
                self.z_central = self.wavelength_of(math.floor(self.shape[0] / 2.)-1)

    @staticmethod
    def from_file(filename, hdu_index=None, verbose=False):
        """
        Factory to create a HyperspectralCube from one HDU in a FITS file.
        http://fits.gsfc.nasa.gov/fits_standard.html

        No other file formats are supported at the moment.
        You may specify the index of the HDU you want.
        If you don't, it will try to guess by searching for a HDU with EXTNAME=DATA,
        or it will pick the first HDU in the list that has 3 dimensions.

        filename: string
            An absolute or relative path to the FITS file we want to load.
            The `astropy.io.fits` module is used to open the file.
        hdu_index: integer|None
            Index of the HDU to load. If you set this, guessing is not attempted.
        verbose: boolean
            Should we fill up the log with endless chatter ?

        :rtype: HyperspectralCube
        """
        try:
            hdu_list = fits.open(filename)
        except Exception:
            raise IOError("Could not open provided FITS file '%s'", filename)

        if verbose:
            logHC.disabled=False
        else:
            logHC.disabled=True

        logHC.info("Opening %s.", filename)

        # Find out which HDU we should read the data from
        hdu = header = None
        extnames = [h.header.get('EXTNAME','').lower() for h in hdu_list]

        def _sanity_check_dimensions(header, hdu_index):
            if header.get('NAXIS') != 3:
                logHC.error("HDU dimensions : NAXIS is only #%d not 3", header.get('NAXIS'))
                raise ValueError("Not a cube")

        # 1/ If set, use HDU indexed at hdu_index
        if hdu_index is not None:
            if hdu_index < 0 or hdu_index >= len(hdu_list):
                raise IOError("Specified HDU #%d is not available", hdu_index)
            logHC.warning("Found extention %d for data" % (hdu_index))

            hdu = hdu_list[hdu_index]
            header = hdu.header
            data = hdu.data
            if hdu_index < len(hdu_list):
                logHC.warning("Using next extention %d for Variance"  % (hdu_index+1))
                variance = hdu_list[hdu_index+1].data
            else:
                variance = None
            _sanity_check_dimensions(header, hdu_index)
        else:
            # 2/ Try to find DATA extension HDU
            found_data_hdu = False
            hdu_index = 0
            while not found_data_hdu and hdu_index < len(hdu_list):
                hdu = hdu_list[hdu_index]
                header = hdu.header
                if 'data' in header.get('EXTNAME', '').lower():
                    try:
                        _sanity_check_dimensions(header, hdu_index)
                    except IOError:
                        hdu_index += 1
                    else:
                        found_data_hdu = True
                else:
                    hdu_index += 1

            # 3/ Try any other HDU, starting with Primary
            if not found_data_hdu:
                hdu_index = 0
                while not found_data_hdu and hdu_index < len(hdu_list):
                    hdu = hdu_list[hdu_index]
                    header = hdu.header
                    try:
                        _sanity_check_dimensions(header, hdu_index)
                    except IOError:
                        hdu_index += 1
                    else:
                        found_data_hdu = True

            # 4/ Alert! Exception raised
            if not found_data_hdu:
                raise IOError("Could not find a HDU containing 3 dimensional "
                              "data in file '%s'." % filename)
            else:
                # Zealous sanity check (skipped if python is run with -O or -OO)
                assert hdu is not None and header is not None

                # Collect and numpyfy data
                data = np.array(hdu.data, dtype=float)

                #look for variance
                found_variance_hdu = False
                hdu_index = 0
                if ('variance' in extnames):
                    hdu_index = hdu_list.index_of('STAT')
                    hdu = hdu_list[hdu_index]
                    header = hdu.header
                    found_variance_hdu = True
                if ('stat' in extnames):
                    hdu_index = hdu_list.index_of('STAT')
                    hdu = hdu_list[hdu_index]
                    header = hdu.header
                    found_variance_hdu = True
                if found_variance_hdu:
                    logHC.info('Found Variance data in file %s' % (filename))
                    try:
                        _sanity_check_dimensions(header, hdu_index)
                        variance = np.array(hdu.data, dtype=float)
                    except IOError:
                        raise IOError("Could not find a variance with proper dimensions")
                else:
                    logHC.debug("Variance data not present in file '%s'." % filename)
                    variance = None
                    
        return HyperspectralCube(data=data, header=header, verbose=verbose, filename=filename, variance=variance)

    @staticmethod
    def from_mpdaf(cube, verbose=False):
        """
        Factory to create a cube from a `MPDAF Cube <http://urania1.univ-lyon1.fr/mpdaf/chrome/site/DocCoreLib/user_manual_cube.html>`_.
        """
        header = cube.data_header
        data = cube.data
        variance = cube.var

        return HyperspectralCube(data=data, header=header, verbose=verbose, filename=cube.filename, variance=variance)

    @property
    def shape(self):
        """
        Returns the shape of the data, that is a tuple of the cardinality of each dimension.
        """
        if self.data is None:
            return ()  # behave consistently with numpy.ndarray
        return self.data.shape

    def is_empty(self):
        """
        Is this cube void of any data ?

        return boolean
        """
        return self.data is None

    def has_header(self):
        """
        Does this cube have a header ?

        return boolean
        """
        return self.header is not None

    def patch(self, crval3=None, crpix3=None, cunit3=None, cdelt3=None, ctype3=None, cunit1=None, force=False):
        """
        Patches up the Cube's missing header values.
        Will raise a ValueError if the values are already in the header unless you use the force=.

        crval3: float
            A value for the cube's header's CRVAL3.
        crpix3: float
            A value for the cube's header's CRPIX3.
        cunit3: str
            A value for the cube's header's CUNIT3.
        cdelt3: float
            A value for the cube's header's CDELT3.
        ctype3: float
            A value for the cube's header's CTYPE3.
        cunit1: str
            A value for the cube's header's CUNIT1.
        force: bool
            Set to True if you want to overwrite existing header cards.
        """

        logHC.info('Patching the header with cards provided')

        if not self.has_header():
            self.header = fits.Header()

        if crval3 is not None:
            if force or self.header.get('CRVAL3') is None:
                self.header.set('CRVAL3', crval3)
            else:
                raise ValueError("The cube already has a CRVAL3 header card. Use the force=.")

        if crpix3 is not None:
            if force or self.header.get('CRPIX3') is None:
                self.header.set('CRPIX3', crpix3)
            else:
                raise ValueError("The cube already has a CRPIX3 header card. Use the force=.")

        if cdelt3 is not None:
            if force or self.header.get('CDELT3') is None:
                self.header.set('CDELT3', cdelt3)
                if 'CD3_3' in self.header.keys():
                     self.header.remove('CD3_3')
            else:
                raise ValueError("The cube already has a CDELT3 header card. Use the force=.")

        if cunit3 is not None:
            if force or self.header.get('CUNIT3') is None:
                self.header.set('CUNIT3', cunit3)
            else:
                raise ValueError("The cube already has a CUNIT3 header card. Use the force=.")

        if ctype3 is not None:
            if force or self.header.get('CTYPE3') is None:
                self.header.set('CTYPE3', ctype3)
            else:
                raise ValueError("The cube already has a CTYPE3 header card. Use the force=.")

        if cunit1 is not None:
            if force or self.header.get('CUNIT1') is None:
                self.header.set('CUNIT1', cunit1)
            else:
                raise ValueError("The cube already has a CUNIT1 header card. Use the force=.")

        #logging.debug('Header after patch ' +  self.__str__() )

    def __str__(self):
        """
        Simple but useful string representation of the header and the data's shape.
        """
        if self.has_header():
            header = self.header.tostring("\n        ").strip()
        else:
            header = "None"
        if self.is_empty():
            data = "None"
        else:
            data = "ndarray of shape %s" % str(self.data.shape)

        return """HyperspectralCube
    header :
        {header}
    data : {data}
""".format(header=header, data=data)

    def __add__(self, other):
        """
        Addition should work pretty much like numpy's ndarray addition.

        HyperspectralCube + Number = HyperspectralCube
            => add the number to each voxel of the cube

        HyperspectralCube + HyperspectralCube = HyperspectralCube
            => requires the input cubes to be of broadcast-compatible shapes

        Raises TypeErrors when cubes are empty or operands are not compatible.
        """
        if self.is_empty():
            raise TypeError("Cannot use operand + on empty HyperspectralCube.")

        if isinstance(other, HyperspectralCube):
            if other.is_empty():
                raise TypeError("Cannot use operand + on empty HyperspectralCube.")
            data = self.data + other.data
        else:
            data = self.data + other  # numpy will raise when inappropriate

        return HyperspectralCube(data=data, header=self.header)

    def __radd__(self, other):
        """
        Addition following Peano axioms is commutative on our sets.
        """
        return self.__add__(other)

    def __sub__(self, other):
        """
        Subtraction should work pretty much like numpy's ndarray subtraction.

        HyperspectralCube - Number = HyperspectralCube
            => subtract the number to each voxel of the cube

        HyperspectralCube - HyperspectralCube = HyperspectralCube
            => requires the input cubes to be of broadcast-compatible shapes

        Raises TypeErrors when cubes are empty or operands are not compatible.
        """
        if self.is_empty():
            raise TypeError("Cannot use operand - on empty HyperspectralCube.")

        if isinstance(other, HyperspectralCube):
            if other.is_empty():
                raise TypeError("Cannot use operand - on empty HyperspectralCube.")
            data = self.data - other.data
        else:
            data = self.data - other  # numpy will raise when inappropriate

        return HyperspectralCube(data=data, header=self.header)

    def __rsub__(self, other):
        """
        Subtraction should work pretty much like numpy's ndarray subtraction.

        Number - HyperspectralCube = HyperspectralCube
            => subtract each voxel of the cube to the number, in each voxel

        HyperspectralCube - HyperspectralCube = HyperspectralCube
            => requires the input cubes to be of broadcast-compatible shapes

        Raises TypeErrors when cubes are empty or operands are not compatible.
        """
        if self.is_empty():
            raise TypeError("Cannot use operand - on empty HyperspectralCube.")

        if isinstance(other, HyperspectralCube):
            if other.is_empty():
                raise TypeError("Cannot use operand - on empty HyperspectralCube.")
            data = other.data - self.data
        else:
            data = other - self.data  # numpy will raise when inappropriate

        return HyperspectralCube(data=data, header=self.header)

    def __mul__(self, other):
        """
        Multiplication should work pretty much like numpy's ndarray multiplication.

        HyperspectralCube * Number = HyperspectralCube
            => multiply each voxel of the cube by the number

        HyperspectralCube * HyperspectralCube = HyperspectralCube
            => requires the input cubes to be of broadcast-compatible shapes

        Raises TypeErrors when cubes are empty or operands are not compatible.
        """
        if self.is_empty():
            raise TypeError("Cannot use operand * on empty HyperspectralCube.")

        if isinstance(other, HyperspectralCube):
            if other.is_empty():
                raise TypeError("Cannot use operand * on empty HyperspectralCube.")
            data = self.data * other.data
        else:
            data = self.data * other  # numpy will raise when inappropriate

        return HyperspectralCube(data=data, header=self.header)

    def __rmul__(self, other):
        """
        Multiplication following Peano axioms is commutative on our sets.
        """
        return self.__mul__(other)

    def __truediv__(self, other):
        """
        Division should work pretty much like numpy's ndarray division.
        We don't check for division by 0, and let exceptions bubble from numpy.
        It's faster, tends to be IEEE 754 compliant, and is consistent with numpy.

        TODO: find out the potential impact of from __future__ import division on this

        HyperspectralCube / Number = HyperspectralCube
            => divide each voxel of the cube by the number

        HyperspectralCube / HyperspectralCube = HyperspectralCube
            => requires the input cubes to be of broadcast-compatible shapes

        Raises TypeErrors when cubes are empty or operands are not compatible.
        """
        if self.is_empty():
            raise TypeError("Cannot use operand / on empty HyperspectralCube.")

        if isinstance(other, HyperspectralCube):
            if other.is_empty():
                raise TypeError("Cannot use operand / on empty HyperspectralCube.")
            data = self.data / other.data
        else:
            data = self.data / other  # numpy will raise when inappropriate

        return HyperspectralCube(data=data, header=self.header)

    def __rdiv__(self, other):
        """
        Division should work pretty much like numpy's ndarray division.
        We don't check for division by 0, and let exceptions bubble from numpy.
        It's faster, tends to be IEEE 754 compliant, and is consistent with numpy.

        Number / HyperspectralCube = HyperspectralCube
            => divide each voxel of the cube by the number

        HyperspectralCube / HyperspectralCube = HyperspectralCube
            => requires the input cubes to be of broadcast-compatible shapes

        Raises TypeErrors when cubes are empty or operands are not compatible.
        """
        if self.is_empty():
            raise TypeError("Cannot use operand / on empty HyperspectralCube.")

        if isinstance(other, HyperspectralCube):
            if other.is_empty():
                raise TypeError("Cannot use operand / on empty HyperspectralCube.")
            data = other.data / self.data
        else:
            data = other / self.data  # numpy will raise when inappropriate

        return HyperspectralCube(data=data, header=self.header)

    def __pow__(self, power):
        """
        Exponentiation should work pretty much like numpy's ndarray exponentiation.

        HyperspectralCube ** Number = HyperspectralCube
            => power each voxel of the cube by the number

        HyperspectralCube ** HyperspectralCube = HyperspectralCube
            => requires the input cubes to be of broadcast-compatible shapes

        Raises TypeErrors when cubes are empty or operands are not compatible.
        """
        if self.is_empty():
            raise TypeError("Cannot use operand ** on empty HyperspectralCube.")

        if isinstance(power, HyperspectralCube):
            if power.is_empty():
                raise TypeError("Cannot use operand ** on empty HyperspectralCube.")
            data = self.data ** power.data
        else:
            data = self.data ** power  # numpy will raise when inappropriate

        return HyperspectralCube(data=data, header=self.header)

    def __rpow__(self, base):
        """
        Exponentiation should work pretty much like numpy's ndarray exponentiation.

        Number ** HyperspectralCube = HyperspectralCube
            => each voxel of the cube is the exponent of the number

        HyperspectralCube ** HyperspectralCube = HyperspectralCube
            => requires the input cubes to be of broadcast-compatible shapes

        Raises TypeErrors when cubes are empty or operands are not compatible.
        """
        if self.is_empty():
            raise TypeError("Cannot use operand ** on empty HyperspectralCube.")

        if isinstance(base, HyperspectralCube):
            if base.is_empty():
                raise TypeError("Cannot use operand ** on empty HyperspectralCube.")
            data = base.data ** self.data
        else:
            data = base ** self.data  # numpy will raise when inappropriate

        return HyperspectralCube(data=data, header=self.header)

    def __getitem__(self, item):
        """
        HyperspectralCube[λmin:λmax, ymin:ymax, xmin:xmax] => HyperspectralCube

        The selector is quite similar to numpy's ndarray selector : min is included, max is excluded.
        You can use negative values, Python slicing syntax, as described here :
        http://docs.scipy.org/doc/numpy/reference/arrays.indexing.html

        :rtype: HyperspectralCube
        """
        if self.is_empty():
            raise TypeError("Cannot use [:,:,:] selector on empty HyperspectralCube.")

        data = self.data[item]

        if self.has_header():
            header = self.header.copy()
            self.check_header()
            # todo : adjust header values to the cut
            raise NotImplementedError("This class does not update the headers. "
                                      "Use mpdaf.obj.Cube to cut AND update headers.")
        else:
            header = None

        return HyperspectralCube(data=data, header=header)

    def __setitem__(self, key, value):
        """
        HyperspectralCube[λmin:λmax, ymin:ymax, xmin:xmax] = HyperspectralCube
        HyperspectralCube[λmin:λmax, ymin:ymax, xmin:xmax] = Number
        HyperspectralCube[λmin:λmax, ymin:ymax, xmin:xmax] = numpy.ndarray

        The indexed mutator is quite similar to numpy's ndarray's : min is included, max is excluded.
        You can use negative values, Python slicing syntax, as described here :
        http://docs.scipy.org/doc/numpy/reference/arrays.indexing.html

        This mutates this HyperspectralCube's data and does not return anything.

        value: HyperspectralCube or numpy.ndarray or Number
            Must be broadcastable to shape defined by the [λ,y,x] slice.

        """
        if self.is_empty():
            raise TypeError("Cannot use [:,:,:] selector on empty HyperspectralCube.")

        if isinstance(value, HyperspectralCube):
            values = value.data
            # fixme: make sure headers are compatible
        else:
            values = value

        self.data[key] = values  # numpy will raise appropriate errors

    def copy(self, out=None):
        """
        Copies this cube into out (if specified) and returns the copy.

        :rtype: HyperspectralCube
        """
        # LBYL pattern, as self.data MUST be a numpy.ndarray and we don't want this to be silenced
        if self.is_empty():
            data = None
        else:
            data = self.data.copy()
            # Same for header, MUST be an instance of astropy.io.fits.Header
        if self.has_header():
            header = self.header.copy()
        else:
            header = None

        if out is None:
            out = HyperspectralCube(data=data, header=header)
        else:
            if not isinstance(out, HyperspectralCube):
                raise IOError("out parameter must be an instance of HyperspectralCube")
            out.data = data
            out.header = data

        return out

    def write_to(self, filename, overwrite=False):
        """
        Write this cube to a FITS file.

        filename: string
            The filename (absolute or relative) of the file we want to write to.
            The `astropy.io.fits` module is used to write to the file.
        overwrite: bool
            When set to True, will overwrite the output file if it exists.
        """
        #astropy v1.2.1 IO does not work with MaskedArrays
        if 'ndarray' in str(type(self.data)):
            primary_hdu = fits.PrimaryHDU(data=self.data, header=self.header)
        elif 'MaskedArray' in str(type(self.data)):
            primary_hdu = fits.PrimaryHDU(data=self.data.data, header=self.header)
        else:
            IOError('Hyperspectral cube data is neither ndarray or MaskedArray')
        hdulist = fits.HDUList([primary_hdu])
        hdulist.writeto(filename, overwrite=overwrite)
        if self.verbose:
            logHC.info("Writing HyperspectralCube to file %s.", filename)
            # Note: astropy already logs something similar, is this really necessary ?

    def get_steps(self):
        """
        Returns a list of the 3 steps [λ,y,x].
        The units are the ones specified in the header.
        """
        if not self.has_header():
            raise IOError("Cannot get the steps of a Cube without header.")
        else:
            #@fixme: need to fix when both present
            if ('CD3_3' in self.header.keys() and 'CDELT3' in self.header.keys()):
                  logHC.info('Both CD3_3 and CDELT3 are present \n will use CDELT3 !')
                  z_step = self.header.remove('CD3_3')
            z_step = self.header.get('CD3_3') or self.header.get('CDELT3')
            try:
                x_step = np.sqrt(self.header.get('CD1_1')**2. + self.header.get('CD1_2')**2. )
                y_step = np.sqrt(self.header.get('CD2_2')**2. + self.header.get('CD2_1')**2. )
            except:
                x_step = self.header.get('CDELT1')
                y_step = self.header.get('CDELT2')


        return [z_step, y_step, x_step]

    def pixel_from_lambda(self, wavelength):
        """
        Returns the pixel index (starting at 0) for the passed wavelength,
        whose unit must be the one specified in the header.

        :rtype: float
        """

        wavelength_array = np.array(wavelength)
        crval = self.header.get('CRVAL3')
        crpix = self.header.get('CRPIX3')
        if 'CD3_3' in self.header.keys() and 'CDELT3' in self.header.keys():
            logHC.info("CD3_3 and CDELT3 are both present\n will use CDELT3")
            self.header.remove('CD3_3')
        cdelt = self.header.get('CDELT3') or self.header.get('CD3_3')

        if crval is None or crpix is None or cdelt is None:
            raise IOError("Cube is missing required 'CRVAL3', 'CRPIX3' or 'CDELT3' header card.")

        return (wavelength_array - crval) / cdelt - 1.0 + crpix

    def wavelength_of(self, pixel_index):
        """
        Get the wavelength (in the unit specified in the header) of the specified pixel index along z.
        The pixel index should start at 0.

        :rtype: float
        """
        pixel_array = np.array(pixel_index)
        crval = self.header.get('CRVAL3')
        crpix = self.header.get('CRPIX3')
        if 'CD3_3' in self.header.keys() and 'CDELT3' in self.header.keys():
            logHC.info("CD3_3 and CDELT3 are both present\n will use CDELT3")
            self.header.remove('CD3_3')
        cdelt = self.header.get('CDELT3') or self.header.get('CD3_3')

        if crval is None or crpix is None or cdelt is None:
            raise IOError("Cube is missing required 'CRVAL3', 'CRPIX3' or 'CDELT3' header card.")

        return (pixel_array - crpix + 1.0) * cdelt + crval

    def sanitize(self, header=None, data=None):
        """
        Procedurally apply various sanitization tasks :
        - http://docs.astropy.org/en/latest/io/fits/usage/verification.html (todo)
        - Fix spectral step keyword :
            - CDEL_3 --> CDELT3
            - CD3_3  --> CDELT3
        - Fix blatantly illegal units :
            - DEG --> deg
            - MICRONS --> um

        Sanitizes this cube's header and data, or provided header and data.

        .. warning::
            header and data are mutated, not copied, so this method returns nothing.
        """

        if header is None:
            header = self.header
        if data is None:
            data = self.data

        if header is None:
            return

        # Fix spectral step keyword
        if header.get('CDELT3') is None:
            if header.get('CDEL_3') is not None:
                header.set('CDELT3', header.get('CDEL_3'))
            if header.get('CD3_3') is not None:
                header.set('CDELT3', header.get('CD3_3'))

        # Fix illegal units
        units_keywords = ['CUNIT1', 'CUNIT2', 'CUNIT3']
        for unit_keyword in units_keywords:
            unit = header.get(unit_keyword)
            if unit is not None:
                unit = unit.strip()
                if unit == 'DEG':
                    unit = 'deg'
                elif 'micron' in unit.lower():
                    unit = 'um'
                header.set(unit_keyword, unit)

    def check_header(self):
        """
        Will check that the header exists and has the cards we need :
            - CUNIT3
            - CDELT3
        We need those cards because they contain calibration markers (start value, step, unit)
        without which the data is pretty much just wasted disk space.

        Raises AttributeError when the header does not check out.

        When you load the data into a HyperspectralCube using a FITS file, the factory will
        require the FITS header to pass this check. (unless you specify check=False)
        Note that the sanitization routine will also automatically create a CDELT3 header card
        from a CDEL_3 header card, for example, and therefore will make this check pass.

        If you want this check to pass, you should add the missing cards to the header using
        assignations like the following (albeit after replacing values in <>) : ::

        self.header.set('<MISSING_CARD>', <missing_value>)
        """

        if self.header is None:
            raise AttributeError("Header is not set. You need a Header for this operation.")

        mandatory_cards = [
            'CUNIT3'
            'CDELT3'
        ]

        for card in mandatory_cards:
            value = self.header.get(card)
            if value is None:
                raise AttributeError("Header's '%s' card is missing, but required for this operation.", card)
